'''
input: 
	smiles batch
 


utility
	1. graph MPN
	2. smiles 
	3. morgan feature 

output:
	1. embedding batch 



deeppurpose
	DDI
	encoders  model 

to do 
	lst -> dataloader -> feature -> model 


	mpnn's feature -> collate -> model 

'''

import csv 
from tqdm import tqdm 
import numpy as np
from copy import deepcopy 
import matplotlib.pyplot as plt

import rdkit
import rdkit.Chem as Chem 
from rdkit import RDLogger  
RDLogger.DisableLog('rdApp.info')
RDLogger.DisableLog('rdApp.*')  
# from rdkit.Chem.EnumerateStereoisomers import EnumerateStereoisomers, StereoEnumerationOptions
import torch 
torch.manual_seed(0)
from torch import nn 
from torch.autograd import Variable
import torch.nn.functional as F
from torch.utils import data  #### data.Dataset 
# device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
from .module import Highway 

def get_drugbank_smiles_lst():
	drugfile = 'data/drugbank_drugs_info.csv'
	with open(drugfile, 'r') as csvfile:
		rows = list(csv.reader(csvfile, delimiter = ','))[1:]
	return [row[27] for row in rows]

def txt_to_lst(text):
	"""
		"['CN[C@H]1CC[C@@H](C2=CC(Cl)=C(Cl)C=C2)C2=CC=CC=C12', 'CNCCC=C1C2=CC=CC=C2CCC2=CC=CC=C12']" 
	"""
	text = text[1:-1]
	lst = [i.strip()[1:-1] for i in text.split(',')]
	return lst 

def get_cooked_data_smiles_lst():
	cooked_file = 'data/raw_data.csv'
	with open(cooked_file, 'r') as csvfile:
		rows = list(csv.reader(csvfile, delimiter = ','))[1:]
	smiles_lst = [row[8] for row in rows]
	smiles_lst = list(map(txt_to_lst, smiles_lst))
	from functools import reduce
	smiles_lst = list(reduce(lambda x,y:x+y, smiles_lst))
	smiles_lst = list(set(smiles_lst))
	# print(len(smiles_lst))  
	return smiles_lst



def create_var(tensor, requires_grad=None):
    if requires_grad is None:
        return Variable(tensor)
    else:
        return Variable(tensor, requires_grad=requires_grad)

def index_select_ND(source, dim, index):
    index_size = index.size()
    suffix_dim = source.size()[1:]
    final_size = index_size + suffix_dim
    target = source.index_select(dim, index.view(-1))
    return target.view(final_size)


def get_mol(smiles):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None: 
        return None
    Chem.Kekulize(mol)
    return mol

ELEM_LIST = ['C', 'N', 'O', 'S', 'F', 'Si', 'P', 'Cl', 'Br', 'Mg', 'Na', 'Ca', 'Fe', 'Al', 'I', 'B', 'K', 'Se', 'Zn', 'H', 'Cu', 'Mn', 'unknown']
ATOM_FDIM = len(ELEM_LIST) + 6 + 5 + 4 + 1
BOND_FDIM = 5 + 6
MAX_NB = 6
### basic setting from https://github.com/wengong-jin/iclr19-graph2graph/blob/master/fast_jtnn/mpn.py

def onek_encoding_unk(x, allowable_set):
    if x not in allowable_set:
        x = allowable_set[-1]
    return list(map(lambda s: x == s, allowable_set))

def atom_features(atom):
    return torch.Tensor(onek_encoding_unk(atom.GetSymbol(), ELEM_LIST) 
            + onek_encoding_unk(atom.GetDegree(), [0,1,2,3,4,5]) 
            + onek_encoding_unk(atom.GetFormalCharge(), [-1,-2,1,2,0])
            + onek_encoding_unk(int(atom.GetChiralTag()), [0,1,2,3])
            + [atom.GetIsAromatic()])

def bond_features(bond):
    bt = bond.GetBondType()
    stereo = int(bond.GetStereo())
    fbond = [bt == Chem.rdchem.BondType.SINGLE, bt == Chem.rdchem.BondType.DOUBLE, bt == Chem.rdchem.BondType.TRIPLE, bt == Chem.rdchem.BondType.AROMATIC, bond.IsInRing()]
    fstereo = onek_encoding_unk(stereo, [0,1,2,3,4,5])
    return torch.Tensor(fbond + fstereo)

def smiles2mpnnfeature(smiles):
	## from mpn.py::tensorize  
	'''
		data-flow:   
			data_process(): apply(smiles2mpnnfeature)
			DBTA: train(): data.DataLoader(data_process_loader())
			mpnn_collate_func()
	'''
	padding = torch.zeros(ATOM_FDIM + BOND_FDIM)
	fatoms, fbonds = [], [padding] 
	in_bonds,all_bonds = [], [(-1,-1)] 
	mol = get_mol(smiles)
	if mol is not None:
		n_atoms = mol.GetNumAtoms()
		for atom in mol.GetAtoms():
			fatoms.append( atom_features(atom))
			in_bonds.append([])

		for bond in mol.GetBonds():
			a1 = bond.GetBeginAtom()
			a2 = bond.GetEndAtom()
			x = a1.GetIdx() 
			y = a2.GetIdx()

			b = len(all_bonds)
			all_bonds.append((x,y))
			fbonds.append( torch.cat([fatoms[x], bond_features(bond)], 0) )
			in_bonds[y].append(b)

			b = len(all_bonds)
			all_bonds.append((y,x))
			fbonds.append( torch.cat([fatoms[y], bond_features(bond)], 0) )
			in_bonds[x].append(b)

		total_bonds = len(all_bonds)
		fatoms = torch.stack(fatoms, 0) 
		fbonds = torch.stack(fbonds, 0) 
		agraph = torch.zeros(n_atoms,MAX_NB).long()
		bgraph = torch.zeros(total_bonds,MAX_NB).long()
		for a in range(n_atoms):
			for i,b in enumerate(in_bonds[a]):
				agraph[a,i] = b

		for b1 in range(1, total_bonds):
			x,y = all_bonds[b1]
			for i,b2 in enumerate(in_bonds[x]):
				if all_bonds[b2][0] != y:
					bgraph[b1,i] = b2
	else: 
		# print('Molecules not found and change to zero vectors..')
		fatoms = torch.zeros(0,39)
		fbonds = torch.zeros(0,50)
		agraph = torch.zeros(0,6)
		bgraph = torch.zeros(0,6)
	Natom, Nbond = fatoms.shape[0], fbonds.shape[0]
	shape_tensor = torch.Tensor([Natom, Nbond]).view(1,-1)
	return [fatoms.float(), fbonds.float(), agraph.float(), bgraph.float(), shape_tensor]


class smiles_dataset(data.Dataset):
	def __init__(self, smiles_lst, label_lst):
		self.smiles_lst = smiles_lst 
		self.label_lst = label_lst 

	def __len__(self):
		return len(self.smiles_lst)

	def __getitem__(self, index):
		smiles = self.smiles_lst[index]
		label = self.label_lst[index]
		smiles_feature = smiles2mpnnfeature(smiles)
		return smiles_feature, label 

## DTI.py --> collate 

## x is a list, len(x)=batch_size, x[i] is tuple, len(x[0])=5  
def mpnn_feature_collate_func(x): 
	return [torch.cat([x[j][i] for j in range(len(x))], 0) for i in range(len(x[0]))]

def mpnn_collate_func(x):
	#print("len(x) is ", len(x)) ## batch_size 
	#print("len(x[0]) is ", len(x[0])) ## 3--- data_process_loader.__getitem__ 
	mpnn_feature = [i[0] for i in x]
	#print("len(mpnn_feature)", len(mpnn_feature), "len(mpnn_feature[0])", len(mpnn_feature[0]))
	mpnn_feature = mpnn_feature_collate_func(mpnn_feature)
	from torch.utils.data.dataloader import default_collate
	x_remain = [i[1:] for i in x]
	x_remain_collated = default_collate(x_remain)
	return [mpnn_feature] + x_remain_collated


def data_loader():
	smiles_lst = get_cooked_data_smiles_lst() 
	label_lst = [1 for i in range(len(smiles_lst))]	
	dataset = smiles_dataset(smiles_lst, label_lst)
	dataloader = data.DataLoader(dataset, batch_size=32, collate_fn = mpnn_collate_func, ) 
	return dataloader 


class MPNN(nn.Sequential):
	def __init__(self, mpnn_hidden_size, mpnn_depth, device):
		super(MPNN, self).__init__()
		self.mpnn_hidden_size = mpnn_hidden_size
		self.mpnn_depth = mpnn_depth 

		self.W_i = nn.Linear(ATOM_FDIM + BOND_FDIM, self.mpnn_hidden_size, bias=False)
		self.W_h = nn.Linear(self.mpnn_hidden_size, self.mpnn_hidden_size, bias=False)
		self.W_o = nn.Linear(ATOM_FDIM + self.mpnn_hidden_size, self.mpnn_hidden_size)

		# initialize weights
		nn.init.xavier_uniform_(self.W_i.weight)
		nn.init.xavier_uniform_(self.W_h.weight)
		nn.init.xavier_uniform_(self.W_o.weight)

		self.device = device
		self = self.to(self.device)

	def set_device(self, device):
		self.device = device 


	@property
	def embedding_size(self):
		return self.mpnn_hidden_size 

	### forward single molecule sequentially. 
	def feature_forward(self, feature):
		''' 
			batch_size == 1 
			feature: utils.smiles2mpnnfeature 
		'''
		fatoms, fbonds, agraph, bgraph, atoms_bonds = feature
		agraph = agraph.long()
		bgraph = bgraph.long()
		#print(fatoms.shape, fbonds.shape, agraph.shape, bgraph.shape, atoms_bonds.shape)
		atoms_bonds = atoms_bonds.long()
		batch_size = atoms_bonds.shape[0]
		N_atoms, N_bonds = 0, 0 
		embeddings = []
		for i in range(batch_size):
			n_a = atoms_bonds[i,0].item()
			n_b = atoms_bonds[i,1].item()
			if (n_a == 0):
				embed = create_var(torch.zeros(1, self.mpnn_hidden_size))
				embeddings.append(embed.to(self.device))
				continue 
			sub_fatoms = fatoms[N_atoms:N_atoms+n_a,:].to(self.device)
			sub_fbonds = fbonds[N_bonds:N_bonds+n_b,:].to(self.device)
			sub_agraph = agraph[N_atoms:N_atoms+n_a,:].to(self.device)
			sub_bgraph = bgraph[N_bonds:N_bonds+n_b,:].to(self.device)
			embed = self.single_feature_forward(sub_fatoms, sub_fbonds, sub_agraph, sub_bgraph)
			embed = embed.to(self.device)           
			embeddings.append(embed)
			N_atoms += n_a
			N_bonds += n_b
		if len(embeddings)==0:
			return None 
		else:
			return torch.cat(embeddings, 0)

	def single_feature_forward(self, fatoms, fbonds, agraph, bgraph):
		'''
			fatoms: (x, 39)
			fbonds: (y, 50)
			agraph: (x, 6)
			bgraph: (y,6)
		'''
		### invalid molecule
		if fatoms.shape[0] == 0:
			return create_var(torch.zeros(1, self.mpnn_hidden_size).to(self.device))
		agraph = agraph.long()
		bgraph = bgraph.long()
		fatoms = create_var(fatoms).to(self.device)
		fbonds = create_var(fbonds).to(self.device)
		agraph = create_var(agraph).to(self.device)
		bgraph = create_var(bgraph).to(self.device)

		binput = self.W_i(fbonds)
		message = F.relu(binput)
		#print("shapes", fbonds.shape, binput.shape, message.shape)
		for i in range(self.mpnn_depth - 1):
			nei_message = index_select_ND(message, 0, bgraph)
			nei_message = nei_message.sum(dim=1)
			nei_message = self.W_h(nei_message)
			message = F.relu(binput + nei_message)

		nei_message = index_select_ND(message, 0, agraph)
		nei_message = nei_message.sum(dim=1)
		ainput = torch.cat([fatoms, nei_message], dim=1)
		atom_hiddens = F.relu(self.W_o(ainput))
		return torch.mean(atom_hiddens, 0).view(1,-1)


	def forward_single_smiles(self, smiles):
		fatoms, fbonds, agraph, bgraph, _ = smiles2mpnnfeature(smiles)
		embed = self.single_feature_forward(fatoms, fbonds, agraph, bgraph).view(1,-1)
		return embed 

	def forward_smiles_lst(self, smiles_lst):
		embed_lst = [self.forward_single_smiles(smiles) for smiles in smiles_lst]
		embed_all = torch.cat(embed_lst, 0)
		return embed_all

	def forward_smiles_lst_average(self, smiles_lst): 
		embed_all = self.forward_smiles_lst(smiles_lst)
		embed_avg = torch.mean(embed_all, 0).view(1,-1)
		return embed_avg


	def forward_smiles_lst_lst(self, smiles_lst_lst): 
		embed_lst = [self.forward_smiles_lst_average(smiles_lst) for smiles_lst in smiles_lst_lst]
		embed_all = torch.cat(embed_lst, 0)  #### n,dim
		return embed_all



class ADMET(nn.Sequential):

	def __init__(self, molecule_encoder, highway_num, device,  
					epoch, lr, weight_decay, save_name):
		super(ADMET, self).__init__()
		self.molecule_encoder = molecule_encoder 
		self.embedding_size = self.molecule_encoder.embedding_size
		self.highway_num = highway_num 
		self.highway_nn_lst = nn.ModuleList([Highway(size = self.embedding_size, num_layers = self.highway_num) for i in range(5)])
		self.fc_output_lst = nn.ModuleList([nn.Linear(self.embedding_size, 1) for i in range(5)])
		self.f = F.relu 
		self.loss = nn.BCEWithLogitsLoss()

		self.epoch = epoch 
		self.lr = lr 
		self.weight_decay = weight_decay 
		self.save_name = save_name 

		self.device = device 
		self = self.to(device)

	def set_device(self, device):
		self.device = device 
		self.molecule_encoder.set_device(device)

	def forward_smiles_lst_embedding(self, smiles_lst, idx):
		embed_all = self.molecule_encoder.forward_smiles_lst(smiles_lst)
		output = self.highway_nn_lst[idx](embed_all)
		return output 

	def forward_embedding_to_pred(self, embeded, idx):
		return self.fc_output_lst[idx](embeded)

	def forward_smiles_lst_pred(self, smiles_lst, idx):
		embeded = self.forward_smiles_lst_embedding(smiles_lst, idx)
		fc_output = self.forward_embedding_to_pred(embeded, idx)
		return fc_output   

	def test(self, dataloader_lst, return_loss = True):
		loss_lst = []
		for idx in range(1):
			single_loss_lst = []
			for smiles_lst, label_vec in dataloader_lst[idx]:
				output = self.forward_smiles_lst_pred(smiles_lst, idx).view(-1)
				loss = self.loss(output, label_vec.to(self.device).float())
				single_loss_lst.append(loss.item())
			loss_lst.append(np.mean(single_loss_lst))
		return np.mean(loss_lst)


if __name__ == "__main__":
	model = MPNN(mpnn_hidden_size = 50, mpnn_depth = 3)
	dataloader = data_loader()
	for smiles_feature, labels in dataloader:
		embedding = model(smiles_feature) 
		print(embedding.shape)
		
	# smiles_lst = get_cooked_data_smiles_lst() 
	# valid_cnt, cnt = 0, 0 
	# for i,smiles in tqdm(enumerate(smiles_lst)):
	# 	feature = smiles2mpnnfeature(smiles)
	# 	if feature is not None:
	# 		valid_cnt += 1
	# 	if i%100==0:
	# 		print("valid rate is", str(valid_cnt/(i+1)))

	### single molecule forward 
	# for smiles in smiles_lst:
	# 	fatoms, fbonds, agraph, bgraph, abshape = smiles2mpnnfeature(smiles)
	# 	embedding = model.single_molecule_forward(fatoms, fbonds, agraph, bgraph)
	# 	print(embedding.shape)














import pdb

import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.nn.utils.rnn import pad_sequence, pack_padded_sequence, pad_packed_sequence
import numpy as np

from pytrial.data.patient_data import SeqPatientCollator
from pytrial.utils.check import (
    check_checkpoint_file, check_model_dir, check_model_config_file, make_dir_if_not_exist
)
from .base import SequenceIndivBase, InputEventEmbedding, RNNModel
from ..data import SequencePatient
from ..losses import XentLoss, BinaryXentLoss, MSELoss, MultilabelBinaryXentLoss
from ..trainer import IndivSeqTrainer


class BuildModel(nn.Module):
    def __init__(self,
        rnn_type,
        emb_size,
        bidirectional,
        vocab_size,
        orders,
        n_rnn_layer,
        output_dim,
        **kwargs,
        ) -> None:
        super().__init__()
        if not isinstance(vocab_size, list): vocab_size = [vocab_size]
        self.rnn = RNNModel(
            rnn_type=rnn_type,
            emb_size=emb_size,
            num_layer=n_rnn_layer,
            bidirectional=bidirectional,
            )
        self.emb_size = emb_size
        self.vocab_size = vocab_size
        self.orders = orders
        self.total_vocab_size = sum(vocab_size)
        self.embedding_matrix = nn.Linear(self.total_vocab_size, emb_size, bias=False)
        if bidirectional:
            self.predictor = nn.Linear(2*emb_size, output_dim)
        else:
            self.predictor = nn.Linear(emb_size, output_dim)

    def forward(self, inputs):
        v = inputs['v']
        v_len = inputs['v_lengths']
        visit_emb = self.embedding_matrix(v) # bs, num_visit, emb_size
        packed_input = pack_padded_sequence(visit_emb, v_len, batch_first=True, enforce_sorted=False)
        packed_output = self.rnn(packed_input)
        output, _ = pad_packed_sequence(packed_output, batch_first=True)
        out_forward = output[range(len(output)), [l - 1 for l in v_len], :self.emb_size]
        out_reverse = output[:, 0, self.emb_size:]
        out_combined = torch.cat((out_forward, out_reverse), 1) # bs, emb_size or emb_size*2(bidirectional)
        return self.predictor(out_combined)

class RNN(SequenceIndivBase):
    '''
    Implement an RNN based model for longitudinal patient records predictive modeling.

    Parameters
    ----------
    vocab_size: list[int]
        A list of vocabulary size for different types of events, e.g., for diagnosis, procedure, medication.

    orders: list[str]
        A list of orders when treating inputs events. Should have the same shape of `vocab_size`.

    output_dim: int
        Output dimension of the model.

        - If binary classification, output_dim=1;
        - If multiclass/multilabel classification, output_dim=n_class
        - If regression, output_dim=1.

    mode: str
        Prediction traget in ['binary','multiclass','multilabel','regression'].

    max_visit: int
        The maximum number of visits for input event codes.

    emb_size: int
        Embedding size for encoding input event codes.

    n_rnn_layer: int
        Number of RNN layers for encoding historical events.

    rnn_type: str
        Pick RNN types in ['rnn','lstm','gru']

    bidirectional: bool
        If True, it encodes historical events in bi-directional manner.
    
    learning_rate: float
        Learning rate for optimization based on SGD. Use torch.optim.Adam by default.

    weigth_decay: float
        Regularization strength for l2 norm; must be a positive float.
        Smaller values specify weaker regularization.
    
    batch_size: int
        Batch size when doing SGD optimization.

    epochs: int
        Maximum number of iterations taken for the solvers to converge.

    evaluation_steps: int
        Number of steps to evaluate the model on validation set or report the training loss.

    num_worker: int
        Number of workers used to do dataloading during training.

    device: str
        The model device.

    experiment_id: str
        The prefix when saving the checkpoints during the training.
    '''
    def __init__(self,
        vocab_size,
        orders,
        mode,
        output_dim=None,
        max_visit=20,
        emb_size=64,
        n_rnn_layer=2,
        rnn_type='lstm',
        bidirectional=False,
        learning_rate=1e-4,
        weight_decay=1e-4,
        batch_size=64,
        epochs=10,
        evaluation_steps=100,
        num_worker=0,
        device='cuda:0',
        experiment_id='test',
        ):
        super().__init__(experiment_id, mode=mode, output_dim=output_dim)
        self.config = {
            'mode':self.mode,
            'vocab_size':vocab_size,
            'max_visit':max_visit,
            'emb_size':emb_size,
            'n_rnn_layer':n_rnn_layer,
            'rnn_type':rnn_type,
            'bidirectional':bidirectional,
            'output_dim':self.output_dim,
            'device':device,
            'learning_rate':learning_rate,
            'batch_size':batch_size,
            'weight_decay':weight_decay,
            'epochs':epochs,
            'num_worker':num_worker,
            'orders':orders,
            'evaluation_steps':evaluation_steps,
            }
        self.config['total_vocab_size'] = sum(vocab_size)
        self.device = device
        self._build_model()

    def fit(self, train_data, valid_data=None):
        '''
        Train model with sequential patient records.
        
        Parameters
        ----------
        train_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events; 'y' corresponds to labels.

        valid_data: SequencePatientBase
            A `SequencePatientBase` contains patient records used to make early stopping of the
            model.
        '''
        self._input_data_check(train_data)
        if valid_data is not None: self._input_data_check(valid_data)
        self._fit_model(train_data, valid_data=valid_data)

    def predict(self, test_data):
        '''
        Predict patient outcomes using longitudinal trial patient sequences.
        
        Parameters
        ----------
        test_data: SequencePatientBase
            A `SequencePatientBase` contains patient records where 'v' corresponds to 
            visit sequence of different events.

        '''
        test_dataloader = self.get_test_dataloader(test_data)
        outputs = self._predict_on_dataloader(test_dataloader)
        return outputs['pred']

    def load_model(self, checkpoint):
        '''
        Load pretrained model from the disk.

        Parameters
        ----------
        checkpoint: str
            The input directory that stores the trained pytorch model and configuration.
        '''
        checkpoint_filename = check_checkpoint_file(checkpoint)
        config_filename = check_model_config_file(checkpoint)
        state_dict = torch.load(checkpoint_filename)
        if config_filename is not None:
            config = self._load_config(config_filename)
            self.config.update(config)
        
        self.model.load_state_dict(state_dict['model'])


    def save_model(self, output_dir):
        '''
        Save the pretrained model to the disk.

        Parameters
        ----------
        output_dir: str
            The output directory that stores the trained pytorch model and configuration.
        '''
        if output_dir is not None:
            make_dir_if_not_exist(output_dir)
        else:
            output_dir = self.checkout_dir
        
        self._save_config(self.config, output_dir=output_dir)
        self._save_checkpoint({'model':self.model.state_dict()}, output_dir=output_dir)

    def get_test_dataloader(self, test_data):
        dataloader = DataLoader(test_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=True,
            shuffle=False,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':test_data.metadata['visit']['mode'],
                    'label_mode':test_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def get_train_dataloader(self, train_data):
        dataloader = DataLoader(train_data,
            batch_size=self.config['batch_size'],
            num_workers=self.config['num_worker'],
            pin_memory=False,
            shuffle=True,
            collate_fn=SeqPatientCollator(
                config={
                    'visit_mode':train_data.metadata['visit']['mode'],
                    'label_mode':train_data.metadata['label']['mode'],
                    }
                ),
            )
        return dataloader

    def _build_model(self):
        self.model = BuildModel(
            rnn_type=self.config['rnn_type'],
            emb_size=self.config['emb_size'],
            max_visit=self.config['max_visit'],
            n_rnn_layer=self.config['n_rnn_layer'],
            bidirectional=self.config['bidirectional'],
            vocab_size=self.config['vocab_size'],
            orders=self.config['orders'],
            output_dim=self.config['output_dim'],
            )
        self.model.to(self.device)

    @torch.no_grad()
    def _predict_on_dataloader(self, test_dataloader):
        pred_list, label_list = [], []
        for batch in test_dataloader:
            inputs = self._prepare_input(batch)
            logits = self.model(inputs)
            pred_list.append(logits)
            if 'y' in batch: label_list.append(batch.pop('y'))
        pred = torch.cat(pred_list, dim=0)
        if self.config['mode'] in ['binary','multilabel']:
            pred = torch.sigmoid(pred)
        if self.config['mode'] == 'multiclass':
            pred = torch.softmax(pred, dim=1)
        pred = pred.cpu().numpy()
        label = torch.cat(label_list) if len(label_list) > 0 else None
        return {'pred':pred,'label':label}

    def _prepare_input(self, data):
        '''
        Prepare inputs for sequential patient record predictive models.

        Parameters
        ----------
        data: dict[list]
            A batch of patient records.

        '''
        visits = data['v']

        if 'x' in data:
            feature = data['x']
            if not isinstance(feature, torch.Tensor): feature = torch.tensor(feature)
            feature = feature.to(self.device)
        else:
            feature = None

        inputs = {
            'v':{},
            'v_lengths':[],
            'x':feature, # baseline feature
            }
        
        v_lengths = [len(visits[self.config['orders'][0]][idx][:self.config['max_visit']]) for idx in range(len(visits[self.config['orders'][0]]))]
        inputs['v_lengths'] = v_lengths

        v = torch.zeros(len(v_lengths), max(v_lengths), self.config['total_vocab_size'])
        for idx in range(len(v_lengths)):
            v[idx,:v_lengths[idx]] = torch.tensor(self._translate_dense_visits_to_sparse({k: visits[k][idx][:self.config['max_visit']] for k in visits}))
        v = v.to(self.device)
        inputs['v'] = v

        if 'y' in data: # target labels
            target = data['y']
            if not isinstance(target, torch.Tensor): target = torch.tensor(target)
            target = target.to(self.device)
            inputs['y'] = target
        return inputs

    def _translate_dense_visits_to_sparse(self, visits):
        total_vocab_size = sum(self.config['vocab_size'])
        num_visits = len(visits[self.config['orders'][0]])
        outputs = np.zeros((num_visits, total_vocab_size))

        for i, o in enumerate(self.config['orders']):
            for j in range(num_visits):
                raw = visits[o][j]
                if isinstance(raw, torch.Tensor): raw = raw.detach().cpu().numpy()
                if i > 0: 
                    voc_size = sum(self.config['vocab_size'][:i-1])
                    if isinstance(raw, list):
                        raw = [r + voc_size for r in raw]
                    else:
                        raw += voc_size
                outputs[j, raw] = 1

        return outputs

    def _fit_model(self, train_data, valid_data=None):
        test_metric_dict = {
            'binary': 'auc',
            'multiclass': 'acc',
            'regression': 'mse',
            'multilabel': 'f1', # take average of F1 scores
            }
        train_dataloader = self.get_train_dataloader(train_data)
        loss_models = self._build_loss_model()
        train_objectives = [(train_dataloader, loss_model) for loss_model in loss_models]

        if self.config['mode'] == 'regression':
            less_is_better = True
        else:
            less_is_better = False

        trainer = IndivSeqTrainer(model=self,
            train_objectives=train_objectives,
            test_data=valid_data,
            test_metric=test_metric_dict[self.config['mode']],
            less_is_better=less_is_better,
            )
        trainer.train(**self.config)

    def _build_loss_model(self):
        mode = self.config['mode']
        if mode == 'binary':
            return [BinaryXentLoss(self.model)]

        if mode == 'multiclass':
            return [XentLoss(self.model)]

        if mode == 'multilabel':
            return [MultilabelBinaryXentLoss(self.model)]

        if mode == 'regression':
            return [MSELoss(self.model)]



import abc
import os

from torch.utils.data import DataLoader

class TrialOutcomeBase(abc.ABC):
	'''Abstract class for all clinical trial outcome prediction

	'''
	@abc.abstractmethod
	def __init__(self):
		pass

	@abc.abstractmethod
	def fit(self, train_data, valid_data):
		raise NotImplementedError

	@abc.abstractmethod
	def predict(self, test_data):
		raise NotImplementedError

	@abc.abstractmethod
	def load_model(self, checkpoint):
		raise NotImplementedError

	@abc.abstractmethod
	def save_model(self, checkpoint):
		raise NotImplementedError
	
	def _build_dataloader_from_dataset(
		self,
		dataset,
		batch_size=64,
		shuffle=False,
		num_workers=0,
		collate_fn=None,
		):
		dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, collate_fn=collate_fn)
		return dataloader




import pdb
import random
import logging
import warnings
from typing import Optional, List

from transformers.models.bart.configuration_bart import BartConfig
from transformers.models.bart.modeling_bart import BartModel, BartPretrainedModel, BartEncoder, BartDecoder
from transformers.models.bart.modeling_bart import shift_tokens_right
from transformers.modeling_outputs import BaseModelOutput, Seq2SeqModelOutput, BaseModelOutputWithPastAndCrossAttentions
import torch
from torch import nn
from torch import Tensor

def _expand_mask(mask: torch.Tensor, dtype: torch.dtype, tgt_len: Optional[int] = None):
    """
    Expands attention_mask from `[bsz, seq_len]` to `[bsz, 1, tgt_seq_len, src_seq_len]`.
    """
    bsz, src_len = mask.size()
    tgt_len = tgt_len if tgt_len is not None else src_len
    expanded_mask = mask[:, None, None, :].expand(bsz, 1, tgt_len, src_len).to(dtype)
    inverted_mask = 1.0 - expanded_mask
    return inverted_mask.masked_fill(inverted_mask.bool(), torch.finfo(dtype).min)

def _all_or_none(values):
    return all(x is None for x in values) or all(x is not None for x in values)

class NumericalConditionalPrompt(nn.Module):
    '''Embedding for conditional prompts based on numerical input patient features,
    take reparametrization trick.

    Parameters
    ----------
    n_feature: number of input features.
    d_model: dimension of output embeddings.
    d_hidden: dimension of intermediate embeddings for reparametrization.
    '''
    def __init__(self, n_feature, d_model, d_hidden) -> None:
        super().__init__()
        self.weight = nn.init.xavier_uniform_(nn.Parameter(Tensor(n_feature, d_hidden)))
        self.bias = nn.init.xavier_uniform_(nn.Parameter(Tensor(n_feature, d_hidden)))
        self.proj = nn.Linear(d_hidden, d_model, bias=False)

    def forward(self, x):
        x = self.weight[None] * x[..., None]
        x = x + self.bias[None]
        x = self.proj(x)
        return x

class CategoricalConditionalPrompt(nn.Module):
    '''Embedding for conditional prompts based on categorical input patient features,
    take reparametrization trick.

    Parameters
    ----------
    cardinalities: the number of distinct values for each feature, e.g., [2, 3, 5] indicates the first cat has 2 possible categories and so on.
    d_model: the output embedding dimension.
    d_hidden: the intermediate layer dimension for reparameterization.
    '''
    def __init__(self,
        cardinalities,
        d_model,
        d_hidden
        ) -> None:
        super().__init__()
        assert cardinalities, 'cardinalities must be non-empty'
        category_offsets = torch.tensor([0] + cardinalities[:-1]).cumsum(0)
        self.register_buffer('category_offsets', category_offsets, persistent=False)
        self.embeddings = nn.Embedding(sum(cardinalities), d_hidden)
        self.bias = nn.init.xavier_uniform_(nn.Parameter(Tensor(len(cardinalities),d_hidden)))
        self.proj = nn.Linear(d_hidden, d_model, bias=False)

    def forward(self, x):
        x = self.embeddings(x + self.category_offsets[None])
        x = x + self.bias[None]
        x = self.proj(x)
        return x

class ConditionalPrompt(nn.Module):
    '''Provide conditional prompt embedding for both categorical and numerical features.

    Parameters
    ----------
    n_num_feature: number of input numerical features.
    cat_cardinalities: a list of unique numbers of each feature.
    d_model: the output dimension.
    d_hidden: the intermediate layer dimension for reparametrization.
    '''
    def __init__(self,
        n_num_feature=None,
        cat_cardinalities=None,
        d_model=None,
        d_hidden=None,
        ) -> None:
        super().__init__()
        if n_num_feature is not None: 
            assert isinstance(n_num_feature, int), 'the passed `n_num_feature` to `promptehr` must be an integer, {} with type {} found.'.format(n_num_feature, type(n_num_feature))
            assert n_num_feature >= 0, 'n_num_feature must be non-negative'
        assert (n_num_feature or cat_cardinalities), 'at least one of n_num_feature or cat_cardinalities must be positive/non-empty'
        self.num_tokenizer = (
            NumericalConditionalPrompt(
                n_feature=n_num_feature,
                d_model=d_model,
                d_hidden=d_hidden,
            )
            if n_num_feature
            else None
        )
        self.cat_tokenizer = (
            CategoricalConditionalPrompt(
                cat_cardinalities,
                d_model=d_model,
                d_hidden=d_hidden,
            )
            if cat_cardinalities
            else None
        )

    def forward(self, x_num=None, x_cat=None):
        '''Perform the forward pass to encode features into prompt context vectors.

        Parameters
        ----------
        x_num: continuous features. Must be presented if :code:`n_num_feature > 0` was passed.
        x_cat: categorical features. Must be presented if non-empty :code:`cat_cardinalities` was passed.
        '''
        assert (
            x_num is not None or x_cat is not None
        ), 'At least one of x_num and x_cat must be presented'
        assert _all_or_none(
            [self.num_tokenizer, x_num]
        ), 'If self.num_tokenizer is (not) None, then x_num must (not) be None'
        assert _all_or_none(
            [self.cat_tokenizer, x_cat]
        ), 'If self.cat_tokenizer is (not) None, then x_cat must (not) be None'
        x = []
        if self.num_tokenizer is not None:
            x.append(self.num_tokenizer(x_num))
        if self.cat_tokenizer is not None:
            x.append(self.cat_tokenizer(x_cat))
        return x[0] if len(x) == 1 else torch.cat(x, dim=1)


class BartLearnedPositionalEmbedding(nn.Embedding):
    """
    This module learns positional embeddings up to a fixed maximum size.
    """

    def __init__(self, num_embeddings: int, embedding_dim: int):
        # Bart is set up so that if padding_idx is specified then offset the embedding ids by 2
        # and adjust num_embeddings appropriately. Other models don't have this hack
        self.offset = 2
        super().__init__(num_embeddings + self.offset, embedding_dim)

    def forward(self, input_ids_shape: torch.Size, past_key_values_length: int = 0):
        """`input_ids_shape` is expected to be [bsz x seqlen]."""
        bsz, seq_len = input_ids_shape[:2]
        positions = torch.arange(
            past_key_values_length, past_key_values_length + seq_len, dtype=torch.long, device=self.weight.device
        )
        positions = positions + self.offset
        positions = torch.minimum(positions, torch.ones_like(positions).to(positions.device)*1024)
        res = super().forward(positions)
        return res

class PromptBartEncoder(BartEncoder):
    def __init__(self, config: BartConfig, embed_tokens: Optional[nn.Embedding] = None):
        super().__init__(config, embed_tokens)
        embed_dim = config.d_model
        self.embed_positions = BartLearnedPositionalEmbedding(
            config.max_position_embeddings,
            embed_dim,
        )

    def forward(self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_prompt_embeds: Optional[torch.FloatTensor]=None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        ):
        '''Make encoding.
        Parameters
        ----------
        inputs_prompt_embeds: Embeded conditional prompt embeddings.
        '''
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # retrieve input_ids and inputs_embeds
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids) * self.embed_scale

        if inputs_prompt_embeds is not None:
            # concatenate prompt embeddings in front of the input embeds
            # modify input_shape and attention_mask at the same time
            inputs_embeds = torch.cat([inputs_prompt_embeds, inputs_embeds], dim=1)
            input_shape = inputs_embeds.size()[:-1]
            if attention_mask is not None:
                add_att_mask = torch.ones(inputs_prompt_embeds.shape[:-1]).to(attention_mask.device)
                attention_mask = torch.cat([add_att_mask, attention_mask], dim=1)

        embed_pos = self.embed_positions(input_shape)
        hidden_states = inputs_embeds + embed_pos
        hidden_states = self.layernorm_embedding(hidden_states)
        hidden_states = nn.functional.dropout(hidden_states, p=self.dropout, training=self.training)

        # expand attention_mask
        if attention_mask is not None:
            # [bsz, seq_len] -> [bsz, 1, tgt_seq_len, src_seq_len]
            attention_mask = _expand_mask(attention_mask, inputs_embeds.dtype)

        encoder_states = () if output_hidden_states else None
        all_attentions = () if output_attentions else None

        # check if head_mask has a correct number of layers specified if desired
        if head_mask is not None:
            if head_mask.size()[0] != (len(self.layers)):
                raise ValueError(
                    f"The head_mask should be specified for {len(self.layers)} layers, but it is for {head_mask.size()[0]}."
                )

        for idx, encoder_layer in enumerate(self.layers):
            if output_hidden_states:
                encoder_states = encoder_states + (hidden_states,)
            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            dropout_probability = random.uniform(0, 1)
            if self.training and (dropout_probability < self.layerdrop):  # skip the layer
                layer_outputs = (None, None)
            else:
                if self.gradient_checkpointing and self.training:

                    def create_custom_forward(module):
                        def custom_forward(*inputs):
                            return module(*inputs, output_attentions)

                        return custom_forward

                    layer_outputs = torch.utils.checkpoint.checkpoint(
                        create_custom_forward(encoder_layer),
                        hidden_states,
                        attention_mask,
                        (head_mask[idx] if head_mask is not None else None),
                    )
                else:
                    layer_outputs = encoder_layer(
                        hidden_states,
                        attention_mask,
                        layer_head_mask=(head_mask[idx] if head_mask is not None else None),
                        output_attentions=output_attentions,
                    )

                hidden_states = layer_outputs[0]

            if output_attentions:
                all_attentions = all_attentions + (layer_outputs[1],)

        if output_hidden_states:
            encoder_states = encoder_states + (hidden_states,)

        if not return_dict:
            return tuple(v for v in [hidden_states, encoder_states, all_attentions] if v is not None)
        return BaseModelOutput(
            last_hidden_state=hidden_states, hidden_states=encoder_states, attentions=all_attentions
        )

class PromptBartDecoder(BartDecoder):
    def __init__(self, config: BartConfig, embed_tokens: Optional[nn.Embedding] = None):
        super().__init__(config, embed_tokens)
        self.embed_positions = BartLearnedPositionalEmbedding(
            config.max_position_embeddings,
            config.d_model,
        )

    def forward(self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_prompt_embeds: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        **kwargs,
        ):
        '''Make forward pass by the decoder.

        Parameters
        ----------
        inputs_prompt_embeds: the embeddings of conditional prompts for the decoder.

        '''
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # retrieve input_ids and inputs_embeds
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both decoder_input_ids and decoder_inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
            input_ids = input_ids.view(-1, input_shape[-1])
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either decoder_input_ids or decoder_inputs_embeds")

        # past_key_values_length
        past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids) * self.embed_scale

        if inputs_prompt_embeds is not None:
            # concatenate prompt embeddings in front of the input embeds
            # modify input_shape and attention_mask at the same time
            inputs_embeds = torch.cat([inputs_prompt_embeds, inputs_embeds], dim=1)
            input_shape = inputs_embeds.size()[:-1]
            if attention_mask is not None:
                add_att_mask = torch.ones(inputs_prompt_embeds.shape[:-1]).to(attention_mask.device)
                attention_mask = torch.cat([add_att_mask, attention_mask], dim=1)

        attention_mask = self._prepare_decoder_attention_mask(
            attention_mask, input_shape, inputs_embeds, past_key_values_length
        )

        # expand encoder attention mask
        if encoder_hidden_states is not None and encoder_attention_mask is not None:
            if inputs_prompt_embeds is not None:
                # adjust for input prompt embeddings
                add_att_mask = torch.ones(inputs_prompt_embeds.shape[:-1]).to(encoder_attention_mask.device)
                encoder_attention_mask = torch.cat([add_att_mask, encoder_attention_mask], dim=1)

            # [bsz, seq_len] -> [bsz, 1, tgt_seq_len, src_seq_len]
            encoder_attention_mask = _expand_mask(encoder_attention_mask, inputs_embeds.dtype, tgt_len=input_shape[-1])

        # embed positions
        positions = self.embed_positions(input_shape, past_key_values_length)

        hidden_states = inputs_embeds + positions
        hidden_states = self.layernorm_embedding(hidden_states)

        hidden_states = nn.functional.dropout(hidden_states, p=self.dropout, training=self.training)

        # decoder layers
        all_hidden_states = () if output_hidden_states else None
        all_self_attns = () if output_attentions else None
        all_cross_attentions = () if (output_attentions and encoder_hidden_states is not None) else None
        next_decoder_cache = () if use_cache else None

        # check if head_mask/cross_attn_head_mask has a correct number of layers specified if desired
        for attn_mask, mask_name in zip([head_mask, cross_attn_head_mask], ["head_mask", "cross_attn_head_mask"]):
            if attn_mask is not None:
                if attn_mask.size()[0] != (len(self.layers)):
                    raise ValueError(
                        f"The `{mask_name}` should be specified for {len(self.layers)} layers, but it is for {head_mask.size()[0]}."
                    )

        for idx, decoder_layer in enumerate(self.layers):
            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            if output_hidden_states:
                all_hidden_states += (hidden_states,)
            dropout_probability = random.uniform(0, 1)
            if self.training and (dropout_probability < self.layerdrop):
                continue

            past_key_value = past_key_values[idx] if past_key_values is not None else None

            if self.gradient_checkpointing and self.training:

                if use_cache:
                    print(
                        "[warning] `use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        # None for past_key_value
                        return module(*inputs, output_attentions, use_cache)

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(decoder_layer),
                    hidden_states,
                    attention_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    head_mask[idx] if head_mask is not None else None,
                    cross_attn_head_mask[idx] if cross_attn_head_mask is not None else None,
                    None,
                )
            else: # testing/generating                
                layer_outputs = decoder_layer(
                    hidden_states,
                    attention_mask=attention_mask,
                    encoder_hidden_states=encoder_hidden_states,
                    encoder_attention_mask=encoder_attention_mask,
                    layer_head_mask=(head_mask[idx] if head_mask is not None else None),
                    cross_attn_layer_head_mask=(
                        cross_attn_head_mask[idx] if cross_attn_head_mask is not None else None
                    ),
                    past_key_value=past_key_value,
                    output_attentions=output_attentions,
                    use_cache=use_cache,
                )
            hidden_states = layer_outputs[0]

            if use_cache:
                next_decoder_cache += (layer_outputs[3 if output_attentions else 1],)

            if output_attentions:
                all_self_attns += (layer_outputs[1],)

                if encoder_hidden_states is not None:
                    all_cross_attentions += (layer_outputs[2],)

        # add hidden states from the last decoder layer
        if output_hidden_states:
            all_hidden_states += (hidden_states,)

        next_cache = next_decoder_cache if use_cache else None
        if not return_dict:
            return tuple(
                v
                for v in [hidden_states, next_cache, all_hidden_states, all_self_attns, all_cross_attentions]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attns,
            cross_attentions=all_cross_attentions,
        )

class PromptBartModel(BartModel):
    '''a subclass of BartModel by using additional prompts for controllable EHR generation.
    '''
    def __init__(self, config: BartConfig):
        super().__init__(config)

        padding_idx, vocab_size = config.pad_token_id, config.vocab_size
        self.shared = nn.Embedding(vocab_size, config.d_model, padding_idx)

        self.encoder = PromptBartEncoder(config, self.shared)
        self.decoder = PromptBartDecoder(config, self.shared)

        # build encoder & decoder prompts
        n_num_feature = config.n_num_feature
        cat_cardinalities = config.cat_cardinalities
        if n_num_feature is not None or cat_cardinalities is not None:
            self.encoder_conditional_prompt = ConditionalPrompt(n_num_feature=n_num_feature,
                cat_cardinalities=cat_cardinalities,
                d_model=config.d_model,
                d_hidden=config.d_prompt_hidden)
            self.decoder_conditional_prompt = ConditionalPrompt(n_num_feature=n_num_feature,
                cat_cardinalities=cat_cardinalities,
                d_model=config.d_model,
                d_hidden=config.d_prompt_hidden)
        else:
            # fix when no baseline feature is provided.
            warnings.warn('No numerical or categorical baseline features are provided, `ConditionalPrompt` is not used in the model.')
            self.encoder_conditional_prompt = None
            self.decoder_conditional_prompt = None

        # Initialize weights and apply final processing
        self.post_init()

    def forward(self,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        x_num: Optional[torch.FloatTensor] = None,
        x_cat: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[List[torch.FloatTensor]] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        ):
        '''Make the forward pass to encode inputs with Bart model.

        Parameters
        ----------
        x_num: the input numerical features, shape (bs, num_feat)
        x_cat: the input categorical features, shape (bs, num_cat)
        '''
        # different to other models, Bart automatically creates decoder_input_ids from
        # input_ids if no decoder_input_ids are provided
        if decoder_input_ids is None and decoder_inputs_embeds is None:
            if input_ids is None:
                raise ValueError(
                    "If no `decoder_input_ids` or `decoder_inputs_embeds` are "
                    "passed, `input_ids` cannot be `None`. Please pass either "
                    "`input_ids` or `decoder_input_ids` or `decoder_inputs_embeds`."
                )

            decoder_input_ids = shift_tokens_right(
                input_ids, self.config.pad_token_id, self.config.decoder_start_token_id
            )

        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            if x_num is not None or x_cat is not None:
                if self.encoder_conditional_prompt is None:
                    warnings.warn('Detect input baseline features in the data,` \
                        but `ConditionalPrompt was not built because no numerical or categorical baseline features are provided when model was initialized. \
                        Consider setting `config.n_num_feature` or `config.cat_cardinalities` when initializing the model.')
                    prompt_embeds = None
                else:
                    prompt_embeds = self.encoder_conditional_prompt(x_num=x_num, x_cat=x_cat)
            else:
                prompt_embeds = None
            
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                head_mask=head_mask,
                inputs_embeds=inputs_embeds,
                inputs_prompt_embeds=prompt_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        # If the user passed a tuple for encoder_outputs, we wrap it in a BaseModelOutput when return_dict=True
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        # decoder outputs consists of (dec_features, past_key_value, dec_hidden, dec_attn)
        if x_num is not None or x_cat is not None:
            if self.decoder_conditional_prompt is None:
                warnings.warn('{} {} {}'.format('Detect input baseline features in the data, but `ConditionalPrompt`',
                    'was not built because no numerical or categorical baseline features',
                    'Consider setting `config.n_num_feature` or `config.cat_cardinalities` when initializing the model.')
                )
                decoder_prompt_embeds = None 
            else:
                decoder_prompt_embeds = self.decoder_conditional_prompt(x_num=x_num, x_cat=x_cat)
        else:
            decoder_prompt_embeds = None
        
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_outputs[0],
            encoder_attention_mask=attention_mask,
            head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            inputs_prompt_embeds=decoder_prompt_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs.last_hidden_state,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )


import pdb
import os
import json
import random
from collections import defaultdict
from typing import Any, Callable, Dict, List, NewType, Optional, Tuple, Union

import numpy as np
import torch
from torch import nn
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset
from transformers import BartTokenizer
from transformers.data.data_collator import InputDataClass

from . import constants

class MimicTrainDataset(Dataset):
    '''mode: for spliting datasets
    '''
    def __init__(self, data_dir:str='./MIMIC-III/processed', mode: str='10k') -> None:
        '''train-5k, 10k, 20k, all
        '''
        self.is_training = True
        # load piece of training data for experiments
        if mode != 'all':
            merge_file = os.path.join(data_dir, f'./MIMIC-III-Merge-train-{mode}.jsonl')
        else:
            merge_file = os.path.join(data_dir, f'./MIMIC-III-Merge-train.jsonl')

        samples = []
        with open(merge_file, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                samples.append(json.loads(line.strip()))
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        sample = self.samples[index]
        return sample

class MimicDataset(Dataset):
    def __init__(self, data_dir:str='./MIMIC-III/processed', mode: str='train') -> None:
        assert mode in ['train', 'test', 'val']
        if mode == 'train': self.is_training = True
        else: self.is_training = False
        # load data
        merge_file = os.path.join(data_dir, f'./MIMIC-III-Merge-{mode}.jsonl')
        samples = []
        with open(merge_file, 'r', encoding='utf-8') as f:
            for line in f.readlines():
                samples.append(json.loads(line.strip()))
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, index):
        sample = self.samples[index]
        return sample

class MimicDataCollator:
    '''Data collator for train/evaluate the EHR-BART model.
    Should keep the whole batch all with features or all without features,
    otherwise raise error!
    '''
    __code_type_list__ = constants.CODE_TYPES
    __special_token_dict__ = constants.SPECIAL_TOKEN_DICT
    __del_or_rep__ = ['rep', 'del']

    def __init__(self, 
        tokenizer,
        code_types,
        n_num_feature,
        mlm_prob=0.15, 
        lambda_poisson=3.0, 
        del_prob=0.15,
        max_train_batch_size=16, 
        drop_feature=False, 
        mode='train'
        ):
        '''mlm_prob: probability of masked tokens
        lambda_poisoon: span infilling parameters
        del_prob: probability of delete tokens
        max_train_batch_size: sample batch to avoid OOM, because for each patient we will generate a batch of series
        '''
        # update code_types
        self.__code_type_list__ = code_types
        self.__special_token_dict__ = {}
        for code in code_types: self.__special_token_dict__[code] = [f'<{code}>', f'</{code}>']

        self.mlm_prob = mlm_prob
        self.tokenizer = tokenizer
        self.tokenizer.model_max_length = constants.model_max_length
        self.mlm_probability = mlm_prob
        self.lambda_poisson = lambda_poisson
        self.del_probability = del_prob
        self.max_train_batch_size = max_train_batch_size # sample batch to avoid OOM
        self.eval_code_type = None # remained for evaluation
        self.drop_feature = drop_feature
        self.n_num_feature = n_num_feature

        assert mode in ['train', 'val', 'test']
        if mode=='train': self.is_training=True
        else: self.is_training=False
        if mode=='test': self.is_testing=True
        else: self.is_testing=False

    def __call__(self, samples: List[InputDataClass]) -> Dict[str, Any]:
        # samples format
        # [{'pid': 'x_num':[], 'x_cat':[], 'diagnosis':[[],[],[],...], 'procedure': [[],[]...], 'drug':[[],[],...] }]
        def _seq_patient_to_promptehr(samples):
            post_samples = []
            for sample in samples:
                post_sample = {}
                visit = sample['v']
                post_sample.update(visit)
                if ('x' in sample) and (self.n_num_feature is not None):
                    if not isinstance(sample['x'], list):
                        sample['x'] = sample['x'].tolist()
                    post_sample['x_num'] = sample['x'][:self.n_num_feature]
                    post_sample['x_cat'] = sample['x'][self.n_num_feature:]
                post_samples.append(post_sample)
            return post_samples
        
        samples = _seq_patient_to_promptehr(samples)

        if self.is_training:
            batch = self.call_train(samples)
        elif self.is_testing:
            batch = self.call_test(samples)
        else:
            batch = self.call_val(samples)
        return batch

    def call_train(self, samples: List[InputDataClass]) -> Dict[str, Any]:
        '''label mask should not be used during training.
        '''
        batch = defaultdict(list)

        # randomly pick one of code types for prediction, keep the same for this batch
        code_type = random.sample(self.__code_type_list__, 1)[0]
        batch['code_type'] = code_type

        for sample in samples:
            num_adm = len(sample[code_type])

            # accumulated during enumerating all admisions
            input_str_all = []
            label_str_all = []
            num_token_all = []

            # cope with too long labtest codes
            # start from the offset if the labtest is too long
            adm = 0
            while adm < num_adm:
                span_str_list = [] # input ids
                span_label_str_list = [] # label ids
                num_token_this_adm = 0

                # shuffle the code order
                code_list = list(sample.keys())
                random.shuffle(code_list)
                for code in sample.keys():
                    if code in ['pid','x_num','x_cat']: continue

                    span = sample[code][adm]

                    if len(span) == 0: continue

                    # restrict the num of tokens in each span
                    span = random.sample(span, min(20, len(span)))

                    # translate span to code_span
                    span = self._process_span(span, code)

                    span_str = self._pad_special_token_head_tail(' '.join(span), code)
                    span_label_str_list.append(span_str)
                    num_token_this_adm += len(span) + 2

                    if code == code_type:
                        # do mask infilling / mask
                        infill_span, _, _ = self.mask_infill([span])
                        span_str = self._pad_special_token_head_tail(' '.join(infill_span[0]), code) 
                        span_str_list.append(span_str)
                    else:
                        if self.__del_or_rep__[random.randint(0,1)] == 'rep': 
                            rep_del_span = self.rep_token([span], code)
                        else: 
                            rep_del_span = self.del_token([span])

                        span_str = self._pad_special_token_head_tail(' '.join(rep_del_span[0]), code) 
                        span_str_list.append(span_str)

                span_str_list.append('</s>')
                span_label_str_list.append('</s>')
                num_token_this_adm += 1

                if adm == 0: # the first visit starts from bos token
                    span_str_list = ['<s>'] + span_str_list
                    span_label_str_list = ['<s>'] + span_label_str_list
                    num_token_this_adm += 1

                # build one sample in batch, concatenate all admissions before
                span_str_this_adm = ' '.join(span_str_list)
                span_label_str_this_adm = ' '.join(span_label_str_list)

                num_token_all, input_str_all, label_str_all = \
                    self._check_max_length(num_token_this_adm, num_token_all, input_str_all, label_str_all)

                # store strs and tokenize at last
                input_strs = ' '.join(input_str_all) + span_str_this_adm
                label_strs = ' '.join(label_str_all) + span_label_str_this_adm
                batch['input_strs'].append(input_strs)
                batch['label_strs'].append(label_strs)

                input_str_all.append(span_str_this_adm)
                label_str_all.append(span_label_str_this_adm)
                num_token_all.append(num_token_this_adm)

                if 'x_num' in sample:
                    batch['x_num'].append(sample['x_num'])
                if 'x_cat' in sample:
                    batch['x_cat'].append(sample['x_cat'])

                if num_adm > 1 and adm < num_adm-1:
                    # build next span prediction task
                    next_span = sample[code_type][adm+1]
                    if len(next_span) == 0:
                        adm +=1 # empty modality, try next admission
                        continue

                    # do shuffling
                    next_span = random.sample(next_span, len(next_span))

                    # translate span to code_span
                    next_span = self._process_span(next_span, code)

                    # do next span prediction
                    label_str = self._pad_special_token_head_tail(' '.join(next_span), code_type)
                    input_str = self._pad_special_token_head_tail('<mask>', code_type)

                    num_token_all, input_str_all, label_str_all = \
                        self._check_max_length(len(next_span)+2, num_token_all, input_str_all, label_str_all)

                    # inputs = self.tokenizer([' '.join(input_str_all)+input_str, ' '.join(label_str_all)+label_str],
                    #     padding='max_length', add_special_tokens=False, return_tensors='pt').input_ids
                    # batch['input_ids'].append(inputs[0].unsqueeze(0))
                    # batch['labels'].append(inputs[1].unsqueeze(0))
                    input_strs = ' '.join(input_str_all)+input_str
                    label_strs = ' '.join(label_str_all)+label_str
                    batch['input_strs'].append(input_strs)
                    batch['label_strs'].append(label_strs)

                    if 'x_num' in sample:
                        batch['x_num'].append(sample['x_num'])
                    if 'x_cat' in sample:
                        batch['x_cat'].append(sample['x_cat'])

                # go to next admission
                adm += 1

        # process all the inputs together
        n_batch = len(batch['input_strs'])
        batch_all_inputs = self.tokenizer(batch.pop('input_strs') + batch.pop('label_strs'), padding=True, add_special_tokens=False, return_tensors='pt')
        batch['input_ids'] = batch_all_inputs['input_ids'][:n_batch]
        batch['attention_mask'] = batch_all_inputs['attention_mask'][:n_batch]
        batch['labels'] = batch_all_inputs['input_ids'][n_batch:]

        if 'x_cat' in batch or 'x_num' in batch:
            # if given features for generation
            if len(batch['x_cat']) < len(batch['input_ids']) or len(batch['x_num']) < len(batch['input_ids']):
                raise ValueError('Features are either not all None or not all assigned for the whole batch, please check the input dataset.')
            else:
                batch['x_cat'] = torch.tensor(batch['x_cat'], dtype=torch.long)
                batch['x_num'] = torch.tensor(batch['x_num'], dtype=torch.float)

        if batch['input_ids'].shape[0] > self.max_train_batch_size:
            # drop some samples if batch size is too large to pass to the model
            sub_indices = np.random.choice(np.arange(len(batch['input_ids'])), self.max_train_batch_size, replace=False)
            batch['input_ids'] = batch['input_ids'][sub_indices]
            batch['attention_mask'] = batch['attention_mask'][sub_indices]
            batch['labels'] = batch['labels'][sub_indices]
            if 'x_cat' in batch or 'x_num' in batch:
                batch['x_cat'] = batch['x_cat'][sub_indices]
                batch['x_num'] = batch['x_num'][sub_indices]

        return batch

    def call_val(self, samples: List[InputDataClass]) -> Dict[str, Any]:
        batch = defaultdict(list)
        eval_code_type = self.eval_code_type
        batch['code_type'] = eval_code_type

        for sample in samples:
            sample_keys = [k for k in sample.keys() if k in self.__code_type_list__]
            num_adm = len(sample[sample_keys[0]])

            # accumulated during enumerating all admisions
            input_str_all = []
            label_str_all = []
            label_mask_list_all = []
            num_token_all = []

            adm = 0
            while adm < num_adm:
                span_str_list = [] # input ids
                span_label_str_list = [] # label ids
                label_mask_list = [] # label mask used for evaluation (not used during training)
                num_token_this_adm = 0

                for code in sample.keys():
                    if code in ['pid','x_num','x_cat']: continue

                    span = sample[code][adm]

                    if len(span) == 0: continue
                    if len(span) > 20: span = random.sample(span, 20)

                    # translate span to code_span
                    span = self._process_span(span, code)

                    num_token_this_adm += len(span) + 2
                    span_str = self._pad_special_token_head_tail(' '.join(span), code) 
                    span_label_str_list.append(span_str)

                    if code == eval_code_type and num_adm == 1:
                        # do mask infilling / mask if there is only one admission of the patient
                        infill_span, label_span, label_mask_span = self.mask_infill([span])
                        span_str = self._pad_special_token_head_tail(' '.join(infill_span[0]), code)
                        span_str_list.append(span_str)
                        label_mask_list += self._pad_label_mask_head_tail(label_mask_span[0]) # add pad token at the head and tail of the label mask

                    else:
                        # do not change anything for not targeted codes
                        span_str = self._pad_special_token_head_tail(' '.join(span), code)
                        span_str_list.append(span_str)
                        label_mask_list += [0] * (len(span)+2) # all masked and the head tail added with zeros

                span_str_list.append('</s>')
                span_label_str_list.append('</s>')
                label_mask_list =  label_mask_list + [0]
                num_token_this_adm += 1

                if adm == 0: # the first visit starts from bos token
                    span_str_list = ['<s>'] + span_str_list
                    span_label_str_list = ['<s>'] + span_label_str_list
                    label_mask_list = [0] + label_mask_list

                # build one sample in batch, concatenate all admissions before
                span_str_this_adm = ' '.join(span_str_list)
                span_label_str_this_adm = ' '.join(span_label_str_list)
                label_mask_this_adm = label_mask_list

                num_token_all, input_str_all, label_str_all, label_mask_list_all = \
                    self._check_max_length(num_token_this_adm, num_token_all, input_str_all, label_str_all, label_mask_list_all)

                if num_adm == 1:
                    batch['input_strs'].append(span_str_this_adm)
                    batch['label_strs'].append(span_label_str_this_adm)
                    batch['label_mask'].append(torch.tensor(label_mask_this_adm))

                    if 'x_num' in sample:
                        batch['x_num'].append(sample['x_num'])
                    if 'x_cat' in sample:
                        batch['x_cat'].append(sample['x_cat'])
                    break # no next span prediction for this

                input_str_all.append(span_str_this_adm)
                label_str_all.append(span_label_str_this_adm)
                label_mask_list_all.append(label_mask_this_adm)
                num_token_all.append(num_token_this_adm)

                if num_adm > 1 and adm < num_adm-1:
                    # build next span predictio ntask
                    next_span = sample[eval_code_type][adm+1]
                    next_span = self._process_span(next_span, eval_code_type)

                    if len(next_span) == 0:
                        adm += 1 # empty modality, try next admission
                        continue # empty modality

                    # do next span prediction
                    label_str = self._pad_special_token_head_tail(' '.join(next_span), eval_code_type)
                    input_str = self._pad_special_token_head_tail('<mask>', eval_code_type)

                    num_token_all, input_str_all, label_str_all, label_mask_list_all = \
                        self._check_max_length(len(next_span)+2, num_token_all, input_str_all, label_str_all, label_mask_list_all)

                    batch['input_strs'].append(' '.join(input_str_all)+input_str)
                    batch['label_strs'].append(' '.join(label_str_all)+label_str)
                    label_mask = sum(label_mask_list_all,[]) + [0] + [1] * len(next_span) + [0]
                    batch['label_mask'].append(torch.tensor(label_mask))

                    if 'x_num' in sample:
                        batch['x_num'].append(sample['x_num'])
                    if 'x_cat' in sample:
                        batch['x_cat'].append(sample['x_cat'])

                # go to next admission
                adm += 1

        if 'x_cat' in batch or 'x_num' in batch:
            # if given features for generation
            if len(batch['x_cat']) < len(batch['input_strs']) or len(batch['x_num']) < len(batch['input_strs']):
                raise ValueError('Features are either not all None or not all assigned for the whole batch, please check the input dataset.')
            else:
                batch['x_cat'] = torch.tensor(batch['x_cat'], dtype=torch.long)
                batch['x_num'] = torch.tensor(batch['x_num'], dtype=torch.float)

        # process all strs inputs at last
        n_batch = len(batch['input_strs'])
        batch_all_inputs = self.tokenizer(batch.pop('input_strs') + batch.pop('label_strs'), padding=True, add_special_tokens=False, return_tensors='pt')
        
        batch['input_ids'] = batch_all_inputs['input_ids'][:n_batch]
        batch['attention_mask'] = batch_all_inputs['attention_mask'][:n_batch]

        batch['labels'] = batch_all_inputs['input_ids'][n_batch:]
        
        # pad all label mask and concat
        batch['label_mask'] = pad_sequence(batch['label_mask'], batch_first=True, padding_value=0)

        # debug: check label mask
        # for i in range(n_batch):
        #     label = batch['labels'][i]
        #     mask = batch['label_mask'][i]
        #     print(self.tokenizer.decode(label[mask==1]))

        return batch

    def call_test(self, samples: List[InputDataClass]) -> Dict[str, Any]:
        '''separate longitudinal and latitudinal perplexity evaluation.
        '''
        batch = defaultdict(list)
        eval_code_type = self.eval_code_type
        eval_ppl_type = self.eval_ppl_type
        batch['code_type'] = eval_code_type

        for sample in samples:
            sample_keys = [k for k in sample.keys() if k in self.__code_type_list__]
            num_adm = len(sample[sample_keys[0]])

            if num_adm == 1 and eval_ppl_type == 'tpl': # cant evaluate tpl if there is only one admission
                continue

            # accumulated during enumerating all admisions
            input_str_all = []
            label_str_all = []
            label_mask_list_all = []
            num_token_all = []

            adm = 0
            while adm < num_adm:
                span_str_list = [] # input ids
                span_label_str_list = [] # label ids
                label_mask_list = [] # label mask used for evaluation (not used during training)
                num_token_this_adm = 0

                for code in sample.keys():
                    if code in ['pid','x_num','x_cat']: continue

                    span = sample[code][adm]

                    if len(span) == 0: continue
                    if len(span) > 20: span = random.sample(span, 20)

                    # translate span to code_span
                    span = self._process_span(span, code)

                    num_token_this_adm += len(span) + 2
                    span_str = self._pad_special_token_head_tail(' '.join(span), code)
                    span_label_str_list.append(span_str)

                    if code == eval_code_type and eval_ppl_type == 'spl': # for spatial ppl evaluation
                        # do mask all codes inside this modality
                        span_str = self._pad_special_token_head_tail('<mask>', code)
                        label_mask_span = [1] * len(span)
                        label_mask_list += self._pad_label_mask_head_tail(label_mask_span)
                        span_str_list.append(span_str)
                    else:
                        # do not change anything for not targeted codes
                        span_str = self._pad_special_token_head_tail(' '.join(span), code)
                        span_str_list.append(span_str)
                        label_mask_list += [0] * (len(span)+2)

                span_str_list.append('</s>')
                span_label_str_list.append('</s>')
                num_token_this_adm += 1
                label_mask_list =  label_mask_list + [0]

                if adm == 0: # the first visit starts from bos token
                    span_str_list = ['<s>'] + span_str_list
                    span_label_str_list = ['<s>'] + span_label_str_list
                    label_mask_list = [0] + label_mask_list

                # build one sample in batch, concatenate all admissions before
                span_str_this_adm = ' '.join(span_str_list)
                span_label_str_this_adm = ' '.join(span_label_str_list)
                label_mask_this_adm = label_mask_list

                input_str_all.append(span_str_this_adm)
                label_str_all.append(span_label_str_this_adm)
                label_mask_list_all.append(label_mask_this_adm)
                num_token_all.append(num_token_this_adm)

                num_token_all, input_str_all, label_str_all, label_mask_list_all = \
                    self._check_max_length(num_token_this_adm, num_token_all, input_str_all, label_str_all, label_mask_list_all)

                if eval_ppl_type == 'spl': # only used when evaluating spatial ppl
                    # if there is only one admission for this patient
                    input_strs = label_str_all[:-1] + [input_str_all[-1]]

                    input_strs = ' '.join(input_strs)
                    label_strs = ' '.join(label_str_all)
                    batch['input_strs'].append(input_strs)
                    batch['label_strs'].append(label_strs)

                    label_mask_past = sum(label_mask_list_all[:-1],[])
                    label_mask = [0] * len(label_mask_past) + label_mask_list_all[-1]
                    batch['label_mask'].append(torch.tensor(label_mask))
                    if 'x_num' in sample:
                        batch['x_num'].append(sample['x_num'])
                    if 'x_cat' in sample:
                        batch['x_cat'].append(sample['x_cat'])

                elif eval_ppl_type == 'tpl' and adm < num_adm-1: # do next span prediction
                    # build next span predictio ntask
                    next_span = sample[eval_code_type][adm+1]
                    next_span = self._process_span(next_span, eval_code_type)

                    if len(next_span) == 0:
                        adm += 1 # empty modality, try next admission
                        continue # empty modality

                    # do next span prediction
                    label_str = self._pad_special_token_head_tail(' '.join(next_span), eval_code_type)
                    input_str = self._pad_special_token_head_tail('<mask>', eval_code_type)
                    
                    num_token_all, input_str_all, label_str_all, label_mask_list_all = \
                        self._check_max_length(len(next_span)+2, num_token_all, input_str_all, label_str_all, label_mask_list_all)

                    label_mask = sum(label_mask_list_all,[]) + [0] + np.ones(len(next_span), dtype=int).tolist() + [0]
                    batch['label_mask'].append(torch.tensor(label_mask))
                    batch['input_strs'].append(' '.join(label_str_all)+input_str)
                    batch['label_strs'].append(' '.join(label_str_all)+label_str)

                    if 'x_num' in sample:
                        batch['x_num'].append(sample['x_num'])
                    if 'x_cat' in sample:
                        batch['x_cat'].append(sample['x_cat'])

                # go to next admission
                adm += 1

        if len(batch['input_strs']) == 0: # num_adm > 1 not found
            return None

        if 'x_cat' in batch or 'x_num' in batch:
            # if given features for generation
            if len(batch['x_cat']) < len(batch['input_strs']) or len(batch['x_num']) < len(batch['input_strs']):
                raise ValueError('Features are either not all None or not all assigned for the whole batch, please check the input dataset.')
            else:
                batch['x_cat'] = torch.tensor(batch['x_cat'], dtype=torch.long)
                batch['x_num'] = torch.tensor(batch['x_num'], dtype=torch.float)

        if self.drop_feature:
            # do not take conditional generation
            batch.pop('x_cat')
            batch.pop('x_num')

        # process all together
        n_batch = len(batch['input_strs'])
        batch_all_inputs = self.tokenizer(batch.pop('input_strs') + batch.pop('label_strs'), padding=True, add_special_tokens=False, return_tensors='pt')
        batch['input_ids'] = batch_all_inputs['input_ids'][:n_batch]
        batch['attention_mask'] = batch_all_inputs['attention_mask'][:n_batch]
        batch['labels'] = batch_all_inputs['input_ids'][n_batch:]
        batch['label_mask'] = pad_sequence(batch['label_mask'], batch_first=True, padding_value=0)
        return batch


    def mask_infill(self, spans:List=[['D_536','D_564']]):
        '''infill in a list of spans.
        '''
        num_adm = len(spans)
        mask_token = self.tokenizer.mask_token
        num_infill_tokens = np.random.poisson(self.lambda_poisson, num_adm)
        label_mask_list = []
        sample_list = []
        label_list = []
        for i, span in enumerate(spans):
            num_code = len(span)
            label_list.append(span)
            if num_code == 1: # only one token inside this span
                label_mask = [1] # 1 -> masked; 0 -> not masked
                sample = ['<mask>']
            else:
                sample = span
                num_infill = num_infill_tokens[i]
                num_infill = max(min(num_code-1, num_infill), 1) # at least mask one token
                label_mask = np.zeros(len(span), dtype=int)
                start_idx = np.random.randint(0, num_infill+1)
                sample = sample[:start_idx] + [mask_token] + sample[start_idx+num_infill:]
                label_mask[start_idx:start_idx+num_infill] = 1
                label_mask = label_mask.tolist()

            sample_list.append(sample)
            label_mask_list.append(label_mask)

        return sample_list, label_list, label_mask_list

    def del_token(self,spans:List=[['D_536', 'D_564']]):
        '''del token
        '''
        return_spans = []
        num_adm = len(spans)
        for i, span in enumerate(spans):
            # span = np.array(span).flatten().tolist()
            num_code = len(span)
            # deletion
            del_indices = np.random.binomial(np.ones(num_code, dtype=int), self.del_probability)
            sub_span = np.array(span)[~del_indices.astype(bool)].tolist()
            return_spans.append(sub_span)

        return return_spans

    def rep_token(self, spans:List=[['D_536', 'D_564']], code_type:str='diag'):
        '''replace token in this span.
        '''
        num_adm = len(spans)
        return_spans = []
        for i, span in enumerate(spans):
            # span = np.array(span).flatten().tolist()
            num_code = len(span)
            # replace
            rep_indices = np.random.binomial(np.ones(num_code, dtype=int), self.del_probability)
            rep_indices = rep_indices.astype(bool)
            random_words = np.random.randint(0, len(self.tokenizer.code_vocab[code_type]), num_code)
            random_words = self.tokenizer.code_vocab[code_type][random_words]
            rep_span = np.array(span).flatten()
            rep_span[rep_indices] = random_words[rep_indices]
            rep_span = rep_span.tolist()
            return_spans.append(rep_span)

        return return_spans

    def set_eval_code_type(self, code_type):
        print(f'evaluation for code {code_type}.')
        self.eval_code_type = code_type

    def set_eval_ppl_type(self, ppl_type):
        assert ppl_type in ['tpl', 'spl'] # temporal or spatial perplexity measure
        print(f'evaluation for {ppl_type} perplexity.')
        self.eval_ppl_type = ppl_type

    def _pad_label_mask_head_tail(self, label_mask_span):
        return [0] + label_mask_span + [0]

    def _pad_special_token_head_tail(self, span, code):
        span_str = self.__special_token_dict__[code][0] + ' ' + span + ' ' + self.__special_token_dict__[code][1]
        return span_str

    def _pad_max_length(self, x:List, fill:int=0):
        '''fill label mask
        '''
        max_length = self.tokenizer.model_max_length
        if len(x) < max_length:
            return x + [fill] * (max_length - len(x))

    def _check_max_length(self, num_token_this_span, num_token_all, input_str_all, label_str_all, label_mask_list_all=None):
        '''cut if it exceeds the model max length
        '''
        while sum(num_token_all) + num_token_this_span > self.tokenizer.model_max_length - 10:
            # print(len(num_token_all))
            # move the span list to the right if the length over passes the maximum length restriction
            num_token_all, input_str_all, label_str_all = num_token_all[1:], input_str_all[1:], label_str_all[1:]
            input_str_all[0] = '<s>' + input_str_all[0]
            label_str_all[0] = '<s>' + label_str_all[0]
            num_token_all[0] += 1
            if label_mask_list_all is not None:
                label_mask_list_all = label_mask_list_all[1:]
                label_mask_list_all[0] = [0] + label_mask_list_all[0]

        if label_mask_list_all is None:
            return num_token_all, input_str_all, label_str_all
        else:
            return num_token_all, input_str_all, label_str_all, label_mask_list_all

    def _process_span(self, span, code):
        return [code+'_'+str(s) for s in span]


import pickle
from copy import deepcopy 
import os
import json
import joblib

from sklearn.metrics import roc_auc_score, f1_score, average_precision_score, precision_score, recall_score, accuracy_score
import matplotlib.pyplot as plt
import numpy as np 
from tqdm import tqdm 
import torch 
from torch import nn 
import torch.nn.functional as F
from functools import reduce

from .model_utils.module import Highway, GCN 
from .model_utils.molecule_encode import MPNN, ADMET 
from .model_utils.icdcode_encode import GRAM, build_icdcode2ancestor_dict
from .model_utils.protocol_encode import Protocol_Embedding
from .model_utils.utils import smiles_lst2fp
from .model_utils.utils import plot_hist
from .model_utils.utils import replace_strange_symbol
from .model_utils.utils import trial_collate_fn
from .base import TrialOutcomeBase

class Interaction(nn.Sequential, TrialOutcomeBase):
    def __init__(self, 
        disease_embedding_dim, 
        protocol_output_dim, 
        molecule_embedding_dim, 
        device, 
        global_embed_size,
        highway_num_layer,
        prefix_name, 
        epoch = 20,
        lr = 3e-4, 
        weight_decay = 0, 
        ):
        super(Interaction, self).__init__()
        icdcode2ancestor_dict = build_icdcode2ancestor_dict()
        self.disease_encoder = GRAM(embedding_dim = disease_embedding_dim, icdcode2ancestor = icdcode2ancestor_dict, device = device)
        self.protocol_encoder = Protocol_Embedding(output_dim = protocol_output_dim, highway_num=3, device = device)
        self.molecule_encoder = MPNN(mpnn_hidden_size = molecule_embedding_dim, mpnn_depth=3, device = device)
        self.global_embed_size = global_embed_size 
        self.highway_num_layer = highway_num_layer 
        self.feature_dim = self.molecule_encoder.embedding_size + self.disease_encoder.embedding_size + self.protocol_encoder.embedding_size
        self.epoch = epoch 
        self.lr = lr 
        self.weight_decay = weight_decay 
        self.save_name = prefix_name + '_interaction'

        self.f = F.relu
        self.loss = nn.BCEWithLogitsLoss()

        ##### NN 
        self.encoder2interaction_fc = nn.Linear(self.feature_dim, self.global_embed_size).to(device)
        self.encoder2interaction_highway = Highway(self.global_embed_size, self.highway_num_layer).to(device)
        self.pred_nn = nn.Linear(self.global_embed_size, 1)

        self.device = device 
        self = self.to(device)

    def feed_lst_of_module(self, input_feature, lst_of_module):
        x = input_feature
        for single_module in lst_of_module:
            x = self.f(single_module(x))
        return x

    def forward_get_three_encoders(self, smiles_lst2, icdcode_lst3, criteria_lst):
        molecule_embed = self.molecule_encoder.forward_smiles_lst_lst(smiles_lst2)
        icd_embed = self.disease_encoder.forward_code_lst3(icdcode_lst3)
        protocol_embed = self.protocol_encoder.forward(criteria_lst)
        return molecule_embed, icd_embed, protocol_embed	

    def forward_encoder_2_interaction(self, molecule_embed, icd_embed, protocol_embed):
        encoder_embedding = torch.cat([molecule_embed, icd_embed, protocol_embed], 1)
        # interaction_embedding = self.feed_lst_of_module(encoder_embedding, [self.encoder2interaction_fc, self.encoder2interaction_highway])
        h = self.encoder2interaction_fc(encoder_embedding)
        h = self.f(h)
        h = self.encoder2interaction_highway(h)
        interaction_embedding = self.f(h)
        return interaction_embedding 

    def forward(self, smiles_lst2, icdcode_lst3, criteria_lst):
        molecule_embed, icd_embed, protocol_embed = self.forward_get_three_encoders(smiles_lst2, icdcode_lst3, criteria_lst)
        interaction_embedding = self.forward_encoder_2_interaction(molecule_embed, icd_embed, protocol_embed)
        output = self.pred_nn(interaction_embedding)
        return output ### 32, 1

    def evaluation(self, predict_all, label_all, threshold = 0.5):
        from sklearn.metrics import roc_curve, precision_recall_curve
        with open("predict_label.txt", 'w') as fout:
            for i,j in zip(predict_all, label_all):
                fout.write(str(i)[:4] + '\t' + str(j)[:4]+'\n')
        auc_score = roc_auc_score(label_all, predict_all)
        figure_folder = "figure"
        #### ROC-curve 
        fpr, tpr, thresholds = roc_curve(label_all, predict_all, pos_label=1)
        # roc_curve =plt.figure()
        # plt.plot(fpr,tpr,'-',label=self.save_name + ' ROC Curve ')
        # plt.legend(fontsize = 15)
        #plt.savefig(os.path.join(figure_folder,name+"_roc_curve.png"))
        #### PR-curve
        precision, recall, thresholds = precision_recall_curve(label_all, predict_all)
        # plt.plot(recall,precision, label = self.save_name + ' PR Curve')
        # plt.legend(fontsize = 15)
        # plt.savefig(os.path.join(figure_folder,self.save_name + "_pr_curve.png"))
        label_all = [int(i) for i in label_all]
        float2binary = lambda x:0 if x<threshold else 1
        predict_all = list(map(float2binary, predict_all))
        f1score = f1_score(label_all, predict_all)
        prauc_score = average_precision_score(label_all, predict_all)
        # print(predict_all)
        precision = precision_score(label_all, predict_all)
        recall = recall_score(label_all, predict_all)
        accuracy = accuracy_score(label_all, predict_all)
        predict_1_ratio = sum(predict_all) / len(predict_all)
        label_1_ratio = sum(label_all) / len(label_all)
        return auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio 

    def testloader_to_lst(self, dataloader):
        nctid_lst, label_lst, smiles_lst2, icdcode_lst3, criteria_lst = [], [], [], [], []
        for nctid, label, smiles, icdcode, criteria in dataloader:
            nctid_lst.extend(nctid)
            label_lst.extend([i.item() for i in label])
            smiles_lst2.extend(smiles)
            icdcode_lst3.extend(icdcode)
            criteria_lst.extend(criteria)
        length = len(nctid_lst)
        assert length == len(smiles_lst2) and length == len(icdcode_lst3)
        return nctid_lst, label_lst, smiles_lst2, icdcode_lst3, criteria_lst, length 

    def generate_predict(self, dataloader):
        whole_loss = 0 
        label_all, predict_all, nctid_all = [], [], []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in dataloader:
            nctid_all.extend(nctid_lst)
            label_vec = label_vec.to(self.device)
            output = self.forward(smiles_lst2, icdcode_lst3, criteria_lst).view(-1)  
            loss = self.loss(output, label_vec.float())
            whole_loss += loss.item()
            predict_all.extend([i.item() for i in torch.sigmoid(output)])
            label_all.extend([i.item() for i in label_vec])

        return whole_loss, predict_all, label_all, nctid_all

    def bootstrap_test(self, dataloader, sample_num = 20):
        # if validloader is not None:
        # 	best_threshold = self.select_threshold_for_binary(validloader)
        self.eval()
        best_threshold = 0.5 
        whole_loss, predict_all, label_all, nctid_all = self.generate_predict(dataloader)
        plt.clf()
        prefix_name = "./figure/" + self.save_name 
        plot_hist(prefix_name, predict_all, label_all)		
        def bootstrap(length, sample_num):
            idx = [i for i in range(length)]
            from random import choices 
            bootstrap_idx = [choices(idx, k = length) for i in range(sample_num)]
            return bootstrap_idx 
        results_lst = []
        bootstrap_idx_lst = bootstrap(len(predict_all), sample_num = sample_num)
        for bootstrap_idx in bootstrap_idx_lst: 
            bootstrap_label = [label_all[idx] for idx in bootstrap_idx]		
            bootstrap_predict = [predict_all[idx] for idx in bootstrap_idx]
            results = self.evaluation(bootstrap_predict, bootstrap_label, threshold = best_threshold)
            results_lst.append(results)
        self.train() 
        auc = [results[0] for results in results_lst]
        f1score = [results[1] for results in results_lst]
        prauc_score = [results[2] for results in results_lst]
        print("PR-AUC   mean: "+str(np.mean(prauc_score))[:6], "std: "+str(np.std(prauc_score))[:6])
        print("F1       mean: "+str(np.mean(f1score))[:6], "std: "+str(np.std(f1score))[:6])
        print("ROC-AUC  mean: "+ str(np.mean(auc))[:6], "std: " + str(np.std(auc))[:6])

        for nctid, label, predict in zip(nctid_all, label_all, predict_all):
            if (predict > 0.5 and label == 0) or (predict < 0.5 and label == 1):
                print(nctid, label, str(predict)[:5])

        nctid2predict = {nctid:predict for nctid, predict in zip(nctid_all, predict_all)} 
        pickle.dump(nctid2predict, open('results/nctid2predict.pkl', 'wb'))
        return nctid_all, predict_all 

    def ongoing_test(self, dataloader, sample_num = 20):
        self.eval()
        best_threshold = 0.5 
        whole_loss, predict_all, label_all, nctid_all = self.generate_predict(dataloader) 
        self.train() 
        return nctid_all, predict_all

    def test(self, dataloader, return_loss = True, validloader=None):
        # if validloader is not None:
        # 	best_threshold = self.select_threshold_for_binary(validloader)
        self.eval()
        best_threshold = 0.5 
        whole_loss, predict_all, label_all, nctid_all = self.generate_predict(dataloader)
        # from HINT.utils import plot_hist
        # plt.clf()
        # prefix_name = "./figure/" + self.save_name 
        # plot_hist(prefix_name, predict_all, label_all)
        self.train()
        if return_loss:
            return whole_loss
        else:
            print_num = 5
            auc_score, f1score, prauc_score, precision, recall, accuracy, \
            predict_1_ratio, label_1_ratio = self.evaluation(predict_all, label_all, threshold = best_threshold)
            print("ROC AUC: " + str(auc_score)[:print_num] + "\nF1: " + str(f1score)[:print_num] \
                 + "\nPR-AUC: " + str(prauc_score)[:print_num] \
                 + "\nPrecision: " + str(precision)[:print_num] \
                 + "\nrecall: "+str(recall)[:print_num] + "\naccuracy: "+str(accuracy)[:print_num] \
                 + "\npredict 1 ratio: " + str(predict_1_ratio)[:print_num] \
                 + "\nlabel 1 ratio: " + str(label_1_ratio)[:print_num])
            return auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio

    def plot_learning_curve(self, train_loss_record, valid_loss_record):
        plt.plot(train_loss_record)
        plt.savefig("./figure/" + self.save_name + '_train_loss.jpg')
        plt.clf() 
        plt.plot(valid_loss_record)
        plt.savefig("./figure/" + self.save_name + '_valid_loss.jpg')
        plt.clf() 

    def select_threshold_for_binary(self, validloader):
        _, prediction, label_all, nctid_all = self.generate_predict(validloader)
        best_f1 = 0
        for threshold in prediction:
            float2binary = lambda x:0 if x<threshold else 1
            predict_all = list(map(float2binary, prediction))
            f1score = precision_score(label_all, predict_all)        
            if f1score > best_f1:
                best_f1 = f1score 
                best_threshold = threshold
        return best_threshold 



class HINT_nograph(Interaction):
    def __init__(self, 
                disease_embedding_dim, 
                protocol_output_dim, 
                molecule_embedding_dim, 		
                device, 
                global_embed_size, 
                highway_num_layer,
                prefix_name, 
                epoch = 20,
                lr = 3e-4, 
                weight_decay = 0, ):
        super(HINT_nograph, self).__init__(					
            disease_embedding_dim = disease_embedding_dim, 
            protocol_output_dim = protocol_output_dim, 
            molecule_embedding_dim = molecule_embedding_dim, 
            device = device,  
            global_embed_size = global_embed_size, 
            prefix_name = prefix_name, 
            highway_num_layer = highway_num_layer,
            epoch = epoch,
            lr = lr, 
            weight_decay = weight_decay, 
            ) 
        self.save_name = prefix_name + '_HINT_nograph'
        '''	### interaction model 
        self.molecule_encoder = molecule_encoder 
        self.disease_encoder = disease_encoder 
        self.protocol_encoder = protocol_encoder 
        self.global_embed_size = global_embed_size 
        self.highway_num_layer = highway_num_layer 
        self.feature_dim = self.molecule_encoder.embedding_size + self.disease_encoder.embedding_size + self.protocol_encoder.embedding_size
        self.epoch = epoch 
        self.lr = lr 
        self.weight_decay = weight_decay 
        self.save_name = save_name

        self.f = F.relu
        self.loss = nn.BCEWithLogitsLoss()

        ##### NN 
        self.encoder2interaction_fc = nn.Linear(self.feature_dim, self.global_embed_size)
        self.encoder2interaction_highway = Highway(self.global_embed_size, self.highway_num_layer)
        self.pred_nn = nn.Linear(self.global_embed_size, 1)
        '''


        #### risk of disease 
        self.risk_disease_fc = nn.Linear(self.disease_encoder.embedding_size, self.global_embed_size)
        self.risk_disease_higway = Highway(self.global_embed_size, self.highway_num_layer)

        #### augment interaction 
        self.augment_interaction_fc = nn.Linear(self.global_embed_size*2, self.global_embed_size)
        self.augment_interaction_highway = Highway(self.global_embed_size, self.highway_num_layer)

        #### ADMET 
        self.admet_model = []
        for i in range(5):
            admet_fc = nn.Linear(self.molecule_encoder.embedding_size, self.global_embed_size).to(device)
            admet_highway = Highway(self.global_embed_size, self.highway_num_layer).to(device)
            self.admet_model.append(nn.ModuleList([admet_fc, admet_highway])) 
        self.admet_model = nn.ModuleList(self.admet_model)

        #### PK 
        self.pk_fc = nn.Linear(self.global_embed_size*5, self.global_embed_size)
        self.pk_highway = Highway(self.global_embed_size, self.highway_num_layer)

        #### trial node 
        self.trial_fc = nn.Linear(self.global_embed_size*2, self.global_embed_size)
        self.trial_highway = Highway(self.global_embed_size, self.highway_num_layer)

        ## self.pred_nn = nn.Linear(self.global_embed_size, 1)

        self.device = device 
        self = self.to(device)


    def forward(self, smiles_lst2, icdcode_lst3, criteria_lst, if_gnn = False):
        ### encoder for molecule, disease and protocol
        molecule_embed, icd_embed, protocol_embed = self.forward_get_three_encoders(smiles_lst2, icdcode_lst3, criteria_lst)
        ### interaction 
        interaction_embedding = self.forward_encoder_2_interaction(molecule_embed, icd_embed, protocol_embed)
        ### risk of disease 
        risk_of_disease_embedding = self.feed_lst_of_module(input_feature = icd_embed, 
                                                            lst_of_module = [self.risk_disease_fc, self.risk_disease_higway])
        ### augment interaction   
        augment_interaction_input = torch.cat([interaction_embedding, risk_of_disease_embedding], 1)
        augment_interaction_embedding = self.feed_lst_of_module(input_feature = augment_interaction_input, 
                                                                lst_of_module = [self.augment_interaction_fc, self.augment_interaction_highway])
        ### admet 
        admet_embedding_lst = []
        for idx in range(5):
            admet_embedding = self.feed_lst_of_module(input_feature = molecule_embed, 
                                                      lst_of_module = self.admet_model[idx])
            admet_embedding_lst.append(admet_embedding)
        ### pk 
        pk_input = torch.cat(admet_embedding_lst, 1)
        pk_embedding = self.feed_lst_of_module(input_feature = pk_input, 
                                               lst_of_module = [self.pk_fc, self.pk_highway])
        ### trial 
        trial_input = torch.cat([pk_embedding, augment_interaction_embedding], 1)
        trial_embedding = self.feed_lst_of_module(input_feature = trial_input, 
                                                  lst_of_module = [self.trial_fc, self.trial_highway])
        output = self.pred_nn(trial_embedding)
        if if_gnn == False:
            return output 
        else:
            embedding_lst = [molecule_embed, icd_embed, protocol_embed, interaction_embedding, risk_of_disease_embedding, \
                             augment_interaction_embedding] + admet_embedding_lst + [pk_embedding, trial_embedding]
            return embedding_lst

class HINT(HINT_nograph):
    '''
    Implement Hierarchical Interaction Network (HINT) model for clinical trial outcome prediction [1]_. 

    Parameters
    ----------
    disease_embedding_dim: int
        dimension of disease code embedding, e.g., 50

    protocol_output_dim: int
        dimension of protocol (eligibility criteria) embedding, e.g., 50

    molecule_embedding_dim: int
        dimension of molecule embedding, e.g., 50
 
    global_embed_size: int
        dimension of trial component embedding, e.g., 50

    highway_num_layer: int
        number of highway layers, e.g., 3

    gnn_hidden_size: int 
        dimension of GNN hidden size, e.g., 50

    epoch: int
        epoch number during training, e.g., 5

    lr: float
        learning rate of optimizer (we use Adam) during training, e.g., 3e-4, 
    
    batch_size: int
        batch size during training, e.g., 32

    weight_decay: float
        weight decay coefficient, e.g., 0.

    prefix_name: str
        name of trial phase as prefix name of the model, e.g., `phase_I`, `phase_II`

    device: str or torch.device
        Target device to train the model, as `cuda:0` or `cpu`.

    Notes
    -----
    .. [1] Fu et al. HINT: Hierarchical Interaction Network for Clinical Trial Outcome Prediction. Cell Patterns, 2022. 
    '''
    def __init__(self,
        disease_embedding_dim=50, 
        protocol_output_dim=50, 
        molecule_embedding_dim=50, 
        global_embed_size=50,
        highway_num_layer=3,
        gnn_hidden_size=50,
        epoch = 20,
        lr = 3e-4,
        batch_size = 32,
        weight_decay = 0,
        prefix_name='phase_I',
        device='cuda:0',
        ):
        super(HINT, self).__init__(
            disease_embedding_dim = disease_embedding_dim, 
            protocol_output_dim = protocol_output_dim, 
            molecule_embedding_dim = molecule_embedding_dim, 
            device = device, 
            prefix_name = prefix_name, 
            global_embed_size = global_embed_size,
            highway_num_layer = highway_num_layer,
            epoch = epoch,
            lr = lr, 
            weight_decay = weight_decay)
        self.save_name = prefix_name
        self.gnn_hidden_size = gnn_hidden_size 
        #### GNN
        self.adj = self.generate_adj()          
        self.gnn = GCN(
            nfeat = self.global_embed_size,
            nhid = self.gnn_hidden_size,
            nclass = 1,
            dropout = 0.6,
            init = 'uniform') 
        ### gnn's attention 		
        self.node_size = self.adj.shape[0]
        self.graph_attention_model_mat = nn.ModuleList([nn.ModuleList([self.gnn_attention() if self.adj[i,j]==1 else None for j in range(self.node_size)]) for i in range(self.node_size)])
        self.device = device 
        self = self.to(device)

        self.config = {
            'disease_embedding_dim': disease_embedding_dim,
            'protocol_output_dim': protocol_output_dim,
            'molecule_embedding_dim': molecule_embedding_dim,
            'global_embed_size': global_embed_size,
            'highway_num_layer': highway_num_layer,
            'gnn_hidden_size': gnn_hidden_size,
            'epoch': epoch,
            'lr': lr,
            'batch_size': batch_size,
            'weight_decay': weight_decay,
            'prefix_name': prefix_name,            
            }

    def predict(self, test_data):
        '''
        Make trial outcome prediction for test data.
        
        Parameters
        ----------
        test_data: TrialOutcomeDatasetBase
            Testing data, should be a `TrialOutcomeDatasetBase` object.
        '''
        # build dataloader using test_data
        testloader = self._build_dataloader_from_dataset(test_data, 
            num_workers=0, batch_size=self.config['batch_size'], shuffle=False,
            collate_fn=trial_collate_fn)
        
        self.eval()
        # best_threshold = 0.5
        whole_loss, predict_all, label_all, nctid_all = self.generate_predict(testloader)
        predict_result = list(zip(nctid_all, predict_all))
        self.train()
        return predict_result

    def fit(self, train_data, valid_data=None):
        '''
        Train HINT model to predict clinical trial outcome (approval rate)

        Parameters
        ----------
        train_data: TrialOutcomeDatasetBase
            Training data, should be a `TrialOutcomeDatasetBase` object.
        
        valid_data: TrialOutcomeDatasetBase
            Validation data, should be a `TrialOutcomeDatasetBase` object. 
        '''
        # build dataloader using train_data
        train_loader = self._build_dataloader_from_dataset(train_data, 
            num_workers=0, batch_size=self.config['batch_size'], shuffle=True,
            collate_fn=trial_collate_fn)

        if valid_data is not None:
            valid_loader = self._build_dataloader_from_dataset(valid_data, 
                num_workers=0, batch_size=self.config['batch_size'], shuffle=False,
                collate_fn=trial_collate_fn)

        opt = torch.optim.Adam(self.parameters(), lr = self.lr, weight_decay = self.weight_decay)
        train_loss_record = [] 
        valid_loss = self.test(valid_loader, return_loss=True)
        valid_loss_record = [valid_loss]
        best_valid_loss = valid_loss
        best_model = deepcopy(self)
        for ep in tqdm(range(self.epoch)):
            for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in tqdm(train_loader):
                label_vec = label_vec.to(self.device)
                output = self.forward(smiles_lst2, icdcode_lst3, criteria_lst).view(-1)  #### 32, 1 -> 32, ||  label_vec 32,
                loss = self.loss(output, label_vec.float())
                train_loss_record.append(loss.item())
                opt.zero_grad() 
                loss.backward() 
                opt.step()
                print('epoch: {}, loss: {}'.format(ep, loss.item()))
            
            if valid_data is not None:
                # only check valid loss when valid_data is not None
                valid_loss = self.test(valid_loader, return_loss=True)
                valid_loss_record.append(valid_loss)
                if valid_loss < best_valid_loss:
                    print('best valid loss: {} -> {}'.format(best_valid_loss, valid_loss))
                    best_valid_loss = valid_loss 
                    best_model = deepcopy(self)

        # self.plot_learning_curve(train_loss_record, valid_loss_record)
        self = deepcopy(best_model)
        # auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio = self.test(test_loader, return_loss = False, validloader = valid_loader)

    def save_model(self, output_dir = None):
        '''
        Save the learned HINT model to the disk.

        Parameters
        ----------
        output_dir: str or None
            The output folder to save the learned model.
            If set None, will save model to `checkpoints/model.ckpt`.
        '''
        if output_dir is None:
            output_dir = 'checkpoints'
        if not os.path.exists(output_dir): os.makedirs(output_dir)
        filename = os.path.join(output_dir, 'model.pkl')

        # save self using joblib
        joblib.dump(self, filename)

        config_filename = os.path.join(output_dir, 'config.json')
        with open(config_filename, 'w') as f:
            json.dump(self.config, f)

    def load_model(self, checkpoint=None):
        '''
        Load the learned HINT model from the disk.

        Parameters
        ----------
        checkpoint: str
            The checkpoint folder to load the learned model.
            The checkpoint under this folder should be `model.ckpt`.
        '''
        if checkpoint is None:
            ckpt_dir = 'checkpoints'
            checkpoint = os.path.join(ckpt_dir, 'model.pkl')
        else:
            checkpoint = os.path.join(checkpoint, 'model.pkl')
        
        # load model using joblib
        model = joblib.load(checkpoint)

        ckpt_dir = os.path.dirname(checkpoint)
        config_filename = os.path.join(ckpt_dir, 'config.json')
        with open(config_filename, 'r') as f:
            model.config = json.load(f)
        
        # replace self with the loaded object
        self.__dict__.update(model.__dict__)

    @staticmethod
    def from_pretrained(checkpoint=None):
        '''
        Load the learned HINT model from the disk.

        Parameters
        ----------
        checkpoint: str 
            The checkpoint folder to load the learned model.
            The checkpoint under this folder should be `model.ckpt`.
        '''
        if checkpoint is None:
            ckpt_dir = 'checkpoints'
            checkpoint = os.path.join(ckpt_dir, 'model.pkl')
        else:
            checkpoint = os.path.join(checkpoint, 'model.pkl')
        
        # load model using joblib
        self = joblib.load(checkpoint)

        ckpt_dir = os.path.dirname(checkpoint)
        config_filename = os.path.join(ckpt_dir, 'config.json')
        with open(config_filename, 'r') as f:
            self.config = json.load(f)
        return self

    def generate_adj(self):        								
        ##### consistent with HINT_nograph.forward
        lst = ["molecule", "disease", "criteria", 'INTERACTION', 'risk_disease', 'augment_interaction', 'A', 'D', 'M', 'E', 'T', 'PK', "final"]
        edge_lst = [("disease", "molecule"), ("disease", "criteria"), ("molecule", "criteria"), 
                    ("disease", "INTERACTION"), ("molecule", "INTERACTION"),  ("criteria", "INTERACTION"), 
                    ("disease", "risk_disease"), ('risk_disease', 'augment_interaction'), ('INTERACTION', 'augment_interaction'),
                    ("molecule", "A"), ("molecule", "D"), ("molecule", "M"), ("molecule", "E"), ("molecule", "T"),
                    ('A', 'PK'), ('D', 'PK'), ('M', 'PK'), ('E', 'PK'), ('T', 'PK'), 
                    ('augment_interaction', 'final'), ('PK', 'final')]
        adj = torch.zeros(len(lst), len(lst))
        adj = torch.eye(len(lst)) * len(lst)
        num2str = {k:v for k,v in enumerate(lst)}
        str2num = {v:k for k,v in enumerate(lst)}
        for i,j in edge_lst:
            n1,n2 = str2num[i], str2num[j]
            adj[n1,n2] = 1
            adj[n2,n1] = 1
        return adj.to(self.device) 

    def generate_attention_matrx(self, node_feature_mat):
        attention_mat = torch.zeros(self.node_size, self.node_size).to(self.device)
        for i in range(self.node_size):
            for j in range(self.node_size):
                if self.adj[i,j]!=1:
                    continue 
                feature = torch.cat([node_feature_mat[i].view(1,-1), node_feature_mat[j].view(1,-1)], 1)
                attention_model = self.graph_attention_model_mat[i][j]
                attention_mat[i,j] = torch.sigmoid(self.feed_lst_of_module(input_feature=feature, lst_of_module=attention_model))
        return attention_mat 

    ##### self.global_embed_size*2 -> 1 
    def gnn_attention(self):
        highway_nn = Highway(size = self.global_embed_size*2, num_layers = self.highway_num_layer).to(self.device)
        highway_fc = nn.Linear(self.global_embed_size*2, 1).to(self.device)
        return nn.ModuleList([highway_nn, highway_fc])	

    def forward(self, smiles_lst2, icdcode_lst3, criteria_lst, return_attention_matrix = False):
        embedding_lst = HINT_nograph.forward(self, smiles_lst2, icdcode_lst3, criteria_lst, if_gnn = True)
        ### length is 13, each is 32,50
        batch_size = embedding_lst[0].shape[0]
        output_lst = []
        if return_attention_matrix:
            attention_mat_lst = []
        for i in range(batch_size):
            node_feature_lst = [embedding[i].view(1,-1) for embedding in embedding_lst]
            node_feature_mat = torch.cat(node_feature_lst, 0) ### 13, 50 
            attention_mat = self.generate_attention_matrx(node_feature_mat)
            output = self.gnn(node_feature_mat, self.adj * attention_mat)
            output = output[-1].view(1,-1)
            output_lst.append(output)
            if return_attention_matrix:
                attention_mat_lst.append(attention_mat)
        output_mat = torch.cat(output_lst, 0)
        if not return_attention_matrix:
            return output_mat
        else:
            return output_mat, attention_mat_lst

    def init_pretrain(self, admet_model):
        self.molecule_encoder = admet_model.molecule_encoder

    ### generate attention matrix 


def dataloader2Xy(nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst, global_icd):
    ## label_vec: (n,)
    y = label_vec 

    num_icd = len(global_icd)
    fp_lst = [smiles_lst2fp(smiles_lst).reshape(1,-1) for smiles_lst in smiles_lst2]
    fp_mat = np.concatenate(fp_lst, 0)
    # fp_mat = torch.from_numpy(fp_mat)  ### (n,2048)

    icdcode_lst = []
    for lst2 in icdcode_lst3:
        lst = list(reduce(lambda x,y:x+y, lst2))
        lst = [i.split('.')[0] for i in lst]
        lst = set(lst)	
        icd_feature = np.zeros((1,num_icd), np.int32)
        for ele in lst:
            if ele in global_icd:
                idx = global_icd.index(ele)
                icd_feature[0,idx] = 1 
        icdcode_lst.append(icd_feature)
    icdcode_mat = np.concatenate(icdcode_lst, 0)
    X = np.concatenate([fp_mat, icdcode_mat], 1)
    X = torch.from_numpy(X)
    X = X.float()
    # icdcode_mat = torch.from_numpy(icdcode_mat) 

    # X = torch.cat([fp_mat, icdcode_mat], 1)
    return X, y 


class FFNN(nn.Sequential):
    def __init__(self, molecule_dim, diseasecode_dim, 
                    global_icd, 
                    protocol_dim = 0,
                    prefix_name = 'FFNN', 
                    epoch = 10,
                    lr = 3e-4, 
                    weight_decay = 0, 
                    ):
        super(FFNN, self).__init__()
        self.molecule_dim = molecule_dim 
        self.diseasecode_dim = diseasecode_dim 
        self.protocol_dim = protocol_dim 
        self.prefix_name = prefix_name 
        self.epoch = epoch 
        self.lr = lr 
        self.weight_decay = weight_decay 
        self.global_icd = global_icd 
        self.num_icd = len(global_icd)

        self.fc_dims = [self.molecule_dim + self.diseasecode_dim + self.protocol_dim, 2000, 1000, 200, 50, 1]
        self.fc_layers = nn.ModuleList([nn.Linear(v,self.fc_dims[i+1]) for i,v in enumerate(self.fc_dims[:-1])])
        self.loss = nn.BCEWithLogitsLoss()
        self.save_name = prefix_name 

    def forward(self, X):
        for i in range(len(self.fc_layers) - 1):
            fc_layer = self.fc_layers[i]
            X = fc_layer(X)
        last_layer = self.fc_layers[-1]
        pred = F.sigmoid(last_layer(X))
        return pred 


        

    def fit(self, train_loader, valid_loader, test_loader):
        opt = torch.optim.Adam(self.parameters(), lr = self.lr, weight_decay = self.weight_decay)
        train_loss_record = [] 
        valid_loss = self.test(valid_loader, return_loss=True)
        valid_loss_record = [valid_loss]
        best_valid_loss = valid_loss
        best_model = deepcopy(self)

        for ep in tqdm(range(self.epoch)):
            for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in train_loader:
                X, _ = dataloader2Xy(nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst, self.global_icd)
                output = self.forward(X).view(-1)  #### 32, 1 -> 32, ||  label_vec 32,
                loss = self.loss(output, label_vec.float())
                train_loss_record.append(loss.item())
                opt.zero_grad() 
                loss.backward() 
                opt.step()
            valid_loss = self.test(valid_loader, return_loss=True)
            valid_loss_record.append(valid_loss)
            if valid_loss < best_valid_loss:
                best_valid_loss = valid_loss 
                best_model = deepcopy(self)

        self.plot_learning_curve(train_loss_record, valid_loss_record)
        self = deepcopy(best_model)
        auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio = self.test(test_loader, return_loss = False, validloader = valid_loader)



    def evaluation(self, predict_all, label_all, threshold = 0.5):
        import pickle, os
        from sklearn.metrics import roc_curve, precision_recall_curve
        with open("predict_label.txt", 'w') as fout:
            for i,j in zip(predict_all, label_all):
                fout.write(str(i)[:4] + '\t' + str(j)[:4]+'\n')
        auc_score = roc_auc_score(label_all, predict_all)
        figure_folder = "figure"
        #### ROC-curve 
        fpr, tpr, thresholds = roc_curve(label_all, predict_all, pos_label=1)
        # roc_curve =plt.figure()
        # plt.plot(fpr,tpr,'-',label=self.save_name + ' ROC Curve ')
        # plt.legend(fontsize = 15)
        #plt.savefig(os.path.join(figure_folder,name+"_roc_curve.png"))
        #### PR-curve
        precision, recall, thresholds = precision_recall_curve(label_all, predict_all)
        # plt.plot(recall,precision, label = self.save_name + ' PR Curve')
        # plt.legend(fontsize = 15)
        # plt.savefig(os.path.join(figure_folder,self.save_name + "_pr_curve.png"))
        label_all = [int(i) for i in label_all]
        float2binary = lambda x:0 if x<threshold else 1
        predict_all = list(map(float2binary, predict_all))
        f1score = f1_score(label_all, predict_all)
        prauc_score = average_precision_score(label_all, predict_all)
        # print(predict_all)
        precision = precision_score(label_all, predict_all)
        recall = recall_score(label_all, predict_all)
        accuracy = accuracy_score(label_all, predict_all)
        predict_1_ratio = sum(predict_all) / len(predict_all)
        label_1_ratio = sum(label_all) / len(label_all)
        return auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio 


    def generate_predict(self, dataloader):
        whole_loss = 0 
        label_all, predict_all = [], []
        for nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst in dataloader:
            X, _ = dataloader2Xy(nctid_lst, label_vec, smiles_lst2, icdcode_lst3, criteria_lst, self.global_icd) 
            output = self.forward(X).view(-1)  
            loss = self.loss(output, label_vec.float())
            whole_loss += loss.item()
            predict_all.extend([i.item() for i in torch.sigmoid(output)])
            label_all.extend([i.item() for i in label_vec])

        return whole_loss, predict_all, label_all

    def bootstrap_test(self, dataloader, sample_num = 20):
        # if validloader is not None:
        # 	best_threshold = self.select_threshold_for_binary(validloader)
        self.eval()
        best_threshold = 0.5 
        whole_loss, predict_all, label_all = self.generate_predict(dataloader)
        plt.clf()
        prefix_name = "./figure/" + self.save_name 
        plot_hist(prefix_name, predict_all, label_all)		
        def bootstrap(length, sample_num):
            idx = [i for i in range(length)]
            from random import choices 
            bootstrap_idx = [choices(idx, k = length) for i in range(sample_num)]
            return bootstrap_idx 
        results_lst = []
        bootstrap_idx_lst = bootstrap(len(predict_all), sample_num = sample_num)
        for bootstrap_idx in bootstrap_idx_lst: 
            bootstrap_label = [label_all[idx] for idx in bootstrap_idx]		
            bootstrap_predict = [predict_all[idx] for idx in bootstrap_idx]
            results = self.evaluation(bootstrap_predict, bootstrap_label, threshold = best_threshold)
            results_lst.append(results)
        self.train() 
        auc = [results[0] for results in results_lst]
        f1score = [results[1] for results in results_lst]
        prauc_score = [results[2] for results in results_lst]
        print("PR-AUC   mean: "+str(np.mean(prauc_score))[:6], "std: "+str(np.std(prauc_score))[:6])
        print("F1       mean: "+str(np.mean(f1score))[:6], "std: "+str(np.std(f1score))[:6])
        print("ROC-AUC  mean: "+ str(np.mean(auc))[:6], "std: " + str(np.std(auc))[:6])


    def test(self, dataloader, return_loss = True, validloader=None):
        # if validloader is not None:
        # 	best_threshold = self.select_threshold_for_binary(validloader)
        self.eval()
        best_threshold = 0.5 
        whole_loss, predict_all, label_all = self.generate_predict(dataloader)
        # from HINT.utils import plot_hist
        # plt.clf()
        # prefix_name = "./figure/" + self.save_name 
        # plot_hist(prefix_name, predict_all, label_all)
        self.train()
        if return_loss:
            return whole_loss
        else:
            print_num = 5
            auc_score, f1score, prauc_score, precision, recall, accuracy, \
            predict_1_ratio, label_1_ratio = self.evaluation(predict_all, label_all, threshold = best_threshold)
            print("ROC AUC: " + str(auc_score)[:print_num] + "\nF1: " + str(f1score)[:print_num] \
                 + "\nPR-AUC: " + str(prauc_score)[:print_num] \
                 + "\nPrecision: " + str(precision)[:print_num] \
                 + "\nrecall: "+str(recall)[:print_num] + "\naccuracy: "+str(accuracy)[:print_num] \
                 + "\npredict 1 ratio: " + str(predict_1_ratio)[:print_num] \
                 + "\nlabel 1 ratio: " + str(label_1_ratio)[:print_num])
            return auc_score, f1score, prauc_score, precision, recall, accuracy, predict_1_ratio, label_1_ratio 


    def plot_learning_curve(self, train_loss_record, valid_loss_record):
        plt.plot(train_loss_record)
        plt.savefig("./figure/" + self.save_name + '_train_loss.jpg')
        plt.clf() 
        plt.plot(valid_loss_record)
        plt.savefig("./figure/" + self.save_name + '_valid_loss.jpg')
        plt.clf() 




class ADMET(nn.Sequential):
    def __init__(self, mpnn_model, device):
        super(ADMET, self).__init__()
        self.num = 5 
        self.mpnn_model = mpnn_model 
        self.device = device 
        self.mpnn_dim = mpnn_model.mpnn_hidden_size 
        self.admet_model = []
        self.global_embed_size = self.mpnn_dim 
        self.highway_num_layer = 2 
        for i in range(5):
            admet_fc = nn.Linear(self.mpnn_model.mpnn_hidden_size, self.global_embed_size).to(device)
            admet_highway = Highway(self.global_embed_size, self.highway_num_layer).to(device)
            self.admet_model.append(nn.ModuleList([admet_fc, admet_highway]))
        self.admet_model = nn.ModuleList(self.admet_model)

        self.admet_pred = nn.ModuleList([nn.Linear(self.global_embed_size,1).to(device) for i in range(5)])
        self.f = F.relu 

        self.device = device 
        self = self.to(device)

    def feed_lst_of_module(self, input_feature, lst_of_module):
        x = input_feature
        for single_module in lst_of_module:
            x = self.f(single_module(x))
        return x 

    def forward(self, smiles_lst, idx):
        assert idx in list(range(5))
        '''
            xxxxxxxxxxxx
        '''
        embeds = self.mpnn_model.forward_smiles_lst_lst(smiles_lst)
        embeds = self.feed_lst_of_module(embeds, self.admet_model[idx]) 
        output = self.admet_pred[idx](embeds)
        return output 

    def test(self, valid_loader):
        pass 

    # def fit(self, train_loader, valid_loader, idx):
    #     opt = torch.optim.Adam(self.parameters(), lr = self.lr, weight_decay = self.weight_decay)
    #     train_loss_record = [] 
    #     valid_loss = self.test(valid_loader, return_loss=True)
    #     valid_loss_record = [valid_loss]
    #     best_valid_loss = valid_loss
    #     best_model = deepcopy(self)

    #     for ep in tqdm(range(self.epoch)):
    #         for smiles_lst in train_loader:
    #             output = self.forward(smiles_lst).view(-1)  #### 32, 1 -> 32, ||  label_vec 32,
    #             loss = self.loss(output, label_vec.float())
    #             train_loss_record.append(loss.item())
    #             opt.zero_grad() 
    #             loss.backward() 
    #             opt.step()
    #         valid_loss = self.test(valid_loader, return_loss=True)
    #         valid_loss_record.append(valid_loss)
    #         if valid_loss < best_valid_loss:
    #             best_valid_loss = valid_loss 
    #             best_model = deepcopy(self)

    #     self = deepcopy(best_model)









'''
Basic patient data functions.
'''
import warnings
import pdb
import json
import pickle
import dill
from collections import defaultdict, OrderedDict
import torch

from torch.utils.data import Dataset, DataLoader
import numpy as np

from ..utils.tabular_utils import HyperTransformer
from ..utils.tabular_utils import get_transformer

class TabularPatientBase(Dataset):
    '''
    Base dataset class for tabular patient records. Subclass it if additional properties and functions
    are required to add for specific tasks. We make use `rdt`: https://docs.sdv.dev/rdt for transform
    and reverse transform of the tabular data.

    Parameters
    ----------
    df: pd.DataFrame
        The input patient tabular format records.

    metadata: dict
        Contains the meta setups of the input data. It should contain the following keys:

        (1) `sdtypes`: dict, the data types of each column in the input data. The keys are the column
            names and the values are the data types. The data types can be one of the following:
            'numerical', 'categorical', 'datetime', 'boolean'.

        (2) `transformers`: dict, the transformers to be used for each column. The keys are the column
            names and the values are the transformer names. The transformer names can be one in
            https://docs.sdv.dev/rdt/transformers-glossary/browse-transformers.
            In addition, we also support inputting a transformer string name, e.g., {'column1': 'OneHotEncoder'}.


        metadata = {

        'sdtypes': {

        'column1': 'numerical',
        'column2': 'boolean',
        'column3': 'datetime',
        'column4': 'categorical',
        'column5': 'categorical',

        },

        transformers':{

        'column1': rdt.transformers.FloatFormatter(missing_value_replacement='mean'),
        'column2': rdt.transformers.BinaryEncoder(missing_value_replacement='mode'),
        'column3': rdt.transformers.UnixTimestampEncoder(missing_value_replacement='mean'),
        'column4': rdt.transformers.FrequencyEncoder(),
        'column5': None, # will not do any thing to this column if no transformer is specified.
        
        }
        
        } 
        
        It is recommend to provide the metadata of the input tabular data.

        - If no metdata is given, the dataset will automatically detect the `dtypes` of columns and build the corresponding `transformers`.
        - If only `sdtypes` are given, the dataset will detect if there are missing `sdtypes` given and build the `transformers` and dtype automatically.

    transform: bool(default=True)
        Whether or not transform raw self.df by hypertransformer.
        If set False, :code:`self.df` will keep as the same as the passed one.

    Examples
    --------
    >>> from pytrial.data.patient_data import TabularPatientBase
    >>> df = pd.read_csv('tabular_patient.csv', index_col=0)
    >>> # set `transform=True` will replace dataset.df with dataset.df_transformed
    >>> dataset = TabularPatientBase(df, transform=True)
    >>> # transform raw dataframe to numerical tables
    >>> df_transformed = dataset.transform(df)
    >>> # make back transform to the original df
    >>> df_raw = dataset.reverse_transform(df_transformed)
    '''
    def __init__(self, df, metadata=None, transform=True):
        self.df = df
        self.metadata = metadata

        # initialize hypertransformer
        self.ht = HyperTransformer()

        if transform:
            if metadata is None:
                warnings.warn('No metadata provided. Metadata will be automatically '
                            'detected from your data. This process may not be accurate. '
                            'We recommend writing metadata to ensure correct data handling.')
                self.ht.detect_initial_config(df)
                self.metadata = self.ht.get_config()
                self.ht.fit(df)

            else:
                # parse the metadata and update hypertransformer's config
                self._parse_metadata()
            
            # create a mapping from column name before to column name after transformation
            self._create_transformed_col2col()

            # replace data with the transformed one
            self.df = self.transform(df)

    def __getitem__(self, index):
        # TODO: support better indexing
        '''
        Indexing the dataframe stored in tabular patient dataset.

        Parameters
        ----------
        index: int or list[int]
            Retrieve the corresponding rows in the dataset.
        '''
        if isinstance(index, int):
            return self.df.iloc[index:index+1]
        
        elif isinstance(index, list):
            return self.df.iloc[index]

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return f'<pytrial.data.patient_data.TabularPatientBase object> Tabular patient data with {self.df.shape[0]} samples {self.df.shape[1]} features, call `.df` to yield the pd.DataFrame data: \n' + repr(self.df)

    def transform(self, df=None):
        '''
        Transform the input df or the self.df by hypertransformer.
        If transform=True in `__init__`, then you do not need to call this function
        to transform self.df because it was tranformed already.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed by self.ht
        '''

        if df is None:
            return self.ht.transform(self.df)
        else:
            return self.ht.transform(df)

    def reverse_transform(self, df=None):
        '''
        Reverse the input dataframe back to the original format. Return the self.df in the original
        format if `df=None`.

        Parameters
        ----------
        df: pd.DataFrame
            The dataframe to be transformed back to the original format by self.ht.
        '''
        if df is None:
            return self.ht.reverse_transform(self.df)
        else:
            return self.ht.reverse_transform(df)

    def _parse_metadata(self):
        '''
        Parse the passed metadata, cope with the following scnearios:
        (1) only `sdtypes` are given;
        (2) only `transformers` are given;
        (3) only partial `sdtypes` are given;
        (4) only partial `transformers` are given.
        '''
        # parse metadata dict for building the hypertransformer
        metadata = self.metadata
        self.ht.detect_initial_config(self.df, verbose=False)

        if 'transformers' in metadata:
            self.ht.update_transformers(metadata['transformers'])
        
        if 'sdtypes' in metadata:
            self.ht.update_sdtypes(metadata['sdtypes'])

        self.ht.fit(self.df)
        self.metadata.update(self.ht.get_config())

    def _create_transformed_col2col(self):
        # create transformed column id to the original columns
        transformed_col2col = OrderedDict()
        for idx, col in enumerate(self.df.columns):
            col_transformer = self.metadata['transformers'][col]
            transformed_col2col[col] = col_transformer.output_columns
        self.metadata['transformed_col2col'] = transformed_col2col


class SequencePatientBase(Dataset):
    '''
    Load sequential patient inputs for longitudinal patient records generation.

    Parameters
    ----------
    data: dict
        A dict contains patient data in sequence and/or in tabular.

        data = {

        'x': np.ndarray or pd.DataFrame
        
        Static patient features in tabular form, typically those baseline information.

        'v': list or np.ndarray
        
        Patient visit sequence in dense format or in tensor format (depends on the model input requirement.)
            
        - If in dense format, it is like [[c1,c2,c3],[c4,c5],...], with shape [n_patient, NA, NA];

        - If in tensor format, it is like [[0,1,1],[1,1,0],...] (multi-hot encoded), with shape [n_patient, max_num_visit, max_num_event].
        
        'y': np.ndarray or pd.Series

        - Target label for each patient if making risk detection, with shape [n_patient, n_class];

        - Target label for each visit if making some visit-level prediction with shape [n_patient, NA, n_class].

        }
    
    metadata: dict (optional)
        
        A dict contains configuration of input patient data.

        metadata = {

        'voc': dict[Voc]

        Vocabulary contains the event index to the exact event name, has three keys in general:
        'diag', 'med', 'prod', corresponding to diagnosis, medication, and procedure.
        ``Voc`` object should have two functions: `idx2word` and `word2idx`.
        
        'visit': dict[str]
        
        a dict contains the format of input data for processing input visit sequences.

        `visit`: {

        'mode': 'tensor' or 'dense',
        
        'order': list[str] (required when `mode='tensor'`)
        
        },

        'label': dict[str]
        
        a dict contains the format of input data for processing input labels.

        `label`: {

        'mode': 'tensor' or 'dense',

        }
    
        'max_visit': int

        the maximum number of visits considered when building tensor inputs, ignored
        when visit mode is dense.
        
        }

    '''
    visit = None
    feature = None
    label = None
    max_visit = None
    visit_voc_size = None
    visit_order = None

    metadata = {
        'voc': {},
        
        'visit':{
            'mode': 'dense',
            'order': ['diag', 'prod', 'med'],
            },

        'label':{
            'mode': 'tensor',
            },

        'max_visit': 20,
    }

    def __init__(self, data, metadata=None) -> None:
        # parse metadata
        self._parse_metadata(metadata=metadata)

        # get input data
        self._parse_inputdata(inputs=data)

    def __getitem__(self, index):
        return_data = {}
        if self.visit is not None:
            visits = self.visit[index]
            if self.metadata['visit']['mode'] == 'tensor':
                visit_ts = self._dense_visit_to_tensor(visits) # return a dict with keys corresponding to order
                return_data['v'] = visit_ts
            else:
                visit_dict = self._parse_dense_visit_with_order(visits) # return a dict with keys corresponding to otder
                return_data['v'] = visit_dict
        
        if self.feature is not None:
            return_data['x'] = self.feature[index]
            
        if self.label is not None:
            return_data['y'] = self.label[index]
        
        return return_data

    def __len__(self):
        return len(self.visit)

    def _get_voc_size(self):
        order = self.metadata['visit']['order']
        vocs = self.metadata['voc']
        voc_size = []
        for order_ in order:
            voc_size.append(
                len(vocs[order_])
            )
        self.visit_voc_size = voc_size

    def _read_pickle(self, file_loc):
        return dill.load(open(file_loc, 'rb'))
    
    def _parse_metadata(self, metadata):
        if metadata is not None: 
            for k,v in metadata.items():
                if isinstance(v, dict):
                    self.metadata[k].update(v)
                else:
                    self.metadata[k] = v
        metadata = self.metadata

        if 'voc' in metadata:
            voc = metadata['voc']

            if 'diag' in voc: self.diag_voc = voc['diag']
            if 'prod' in voc: self.prod_voc = voc['prod']
            if 'med' in voc: self.med_voc = voc['med']
        
        if metadata['visit']['mode'] == 'tensor':
            self._get_voc_size()
        
        if 'order' in metadata['visit']:
            self.visit_order = metadata['visit']['order']

        if 'max_visit' in metadata:
            self.max_visit = metadata['max_visit']

    def _parse_inputdata(self, inputs):
        if 'x' in inputs: self.feature = inputs['x']
        if 'v' in inputs: self.visit = inputs['v']
        if 'y' in inputs: self.label = inputs['y']

    def _dense_visit_to_tensor(self, visits):
        res = {}
        for i,o in enumerate(self.visit_order):
            res[o] = np.zeros((self.max_visit, self.visit_voc_size[i]), dtype=int)

        for i, visit in enumerate(visits):
            # clip if the max visit is larger than self.max_visit
            if i >= self.max_visit: break
            
            for j, o in enumerate(self.visit_order):
                res[o][i, visit[j]] = 1
        return res
    
    def _parse_dense_visit_with_order(self, visits):
        return_data = defaultdict(list)
        order_list = self.metadata['visit']['order']
        for visit in visits:
            for i, o in enumerate(order_list):
                return_data[o].append(visit[i])
        return return_data
    
    def get_label(self):
        return self.label

class SeqPatientCollator:
    '''
    support make collation of unequal sized list of batch for densely stored
    sequential visits data.

    Parameters
    ----------
    config: dict
        
        {

        'visit_mode': in 'dense' or 'tensor',

        'label_mode': in 'dense' or 'tensor',

        }
        
    '''
    is_tensor_visit = False
    is_tensor_label = False
    config = {'visit_mode':'dense', 'label_mode':'tensor'} 

    def __init__(self, config=None):
        if config is not None:
            self.config.update(config)
            
        if self.config['visit_mode'] == 'tensor': self.is_tensor_visit = True
        else: self.is_tensor_visit = False

        if self.config['label_mode'] == 'tensor': self.is_tensor_label = True
        else: self.is_tensor_label = False

    def __call__(self, inputs):
        '''
        Paramters
        ---------
        inputs = {
            'v': {
                'diag': list[np.ndarray],
                'prod': list[np.ndarray],
                'med': list[np.ndarray],
                },

            'x': list[np.ndarray],

            'y': list[np.ndarray]
        }

        Returns
        -------
        outputs = {
            'v':{ # visit event sequence
                'diag': tensor or list[tensor],
                'prod': tensor or list[tensor],
                'med': tensor or list[tensor],
            },
            'x': tensor, # static features
            'y': tensor or list, # patient-level label in tensor or visit-level label in list[tensor]
        }
        '''
        # init output dict
        return_data = defaultdict(list)
        return_data['v'] = defaultdict(list)

        for input in inputs:
            for k, v in input.items():
                if k == 'v': # visit seq
                    for key, value in v.items():
                        return_data['v'][key].append(value)
                else: # feature and label
                    return_data[k].append(v)

        # processing all
        if self.is_tensor_visit:
            self._parse_tensor_visit(return_data)
        
        if self.is_tensor_label:
            self._parse_tensor_label(return_data)

        if 'x' in input:
            self._parse_tensor_feature(return_data)

        return return_data
    
    def _parse_tensor_visit(self, return_data):
        for k, v in return_data['v'].items():
            if isinstance(v, list):
                v = np.array(v)
            return_data['v'][k] = torch.tensor(v, dtype=int)

    def _parse_tensor_label(self, return_data):
        y_ts = torch.tensor(np.array(return_data['y']))
        return_data['y'] = y_ts

    def _parse_tensor_feature(self, return_data):
        x_ts = torch.tensor(np.array(return_data['x']))
        return_data['x'] = x_ts